from typing import Optional, Union

import numpy as np

from osaft import log
from osaft.core.backgroundfields import WaveTypes
from osaft.core.frequency import Frequency
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import full_range
from osaft.core.geometries import Sphere
from osaft.core.helper import InputHandler
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_scattering import BaseScattering
from osaft.solutions.Doinikov1994Compressible.base import (
    BaseDoinikov1994Compressible,
)
from osaft.solutions.Doinikov1994Compressible.coefficientmatrix import (
    CoefficientMatrix,
)


class ScatteringField(
    CoefficientMatrix,
    BaseDoinikov1994Compressible,
    BaseScattering,
):
    """Scattering field of Doinikov (viscous fluid-viscous sphere; 1994)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the solid [m]
    :param rho_s: Density of the fluid-like solid [kg/m^3]
    :param c_s: Speed of sound of the sphere [m/s]
    :param eta_s: shear viscosity of the sphere [Pa s]
    :param zeta_s: bulk viscosity of the sphere [Pa s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param position: Position within the standing wave field [m]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param N_max: Highest order mode
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float, c_s: float, eta_s: float, zeta_s: float,
        rho_f: float, c_f: float, eta_f: float, zeta_f: float,
        p_0: float, wave_type: WaveTypes, position: float,
        N_max: Optional[int] = 3,
    ) -> None:
        """Constructor method
        """

        # init of parent class
        CoefficientMatrix.__init__(self)
        BaseDoinikov1994Compressible.__init__(
            self, frequency, radius,
            rho_s, c_s, eta_s, zeta_s,
            rho_f, c_f, eta_f, zeta_f,
            p_0, position, wave_type, N_max,
        )

        self._create_variables()
        self._set_links()

        if type(self) is ScatteringField:
            log.debug(repr(self))
            log.info(str(self))

    def _create_variables(self) -> None:
        # Dependent variables
        self._det_M_n = ActiveVariable(
            self._reset_list,
            'determinant of M matrix',
        )
        self._det_M_1_n = ActiveVariable(
            self._reset_list,
            'determinant of M_1 matrix',
        )
        self._det_M_2_n = ActiveVariable(
            self._reset_list,
            'determinant of M_2 matrix',
        )
        self._det_M_3_n = ActiveVariable(
            self._reset_list,
            'determinant of M_3 matrix',
        )
        self._det_M_4_n = ActiveVariable(
            self._reset_list,
            'determinant of M_4 matrix',
        )

        self._alpha_n = ActiveVariable(
            self._reset_list,
            'fluid scattering coefficient alpha_n',
        )

        self._beta_n = ActiveVariable(
            self._reset_list,
            'fluid scattering coefficient beta_n',
        )

        self._alpha_hat_n = ActiveVariable(
            self._reset_list, 'sphere scattering coefficient alpha_hat_n',
        )

        self._beta_hat_n = ActiveVariable(
            self._reset_list, 'sphere scattering coefficient beta_hat_n',
        )

        self._x = ActiveVariable(self._compute_x)
        self._x_v = ActiveVariable(self._compute_x_v)

        self._x_hat = ActiveVariable(self._compute_x_hat)
        self._x_hat_v = ActiveVariable(self._compute_x_hat_v)

    def _set_links(self) -> None:
        # define dependencies
        self._x.is_computed_by(
            self.fluid._k_f, self.sphere._R_0,
        )
        self._x_v.is_computed_by(
            self.fluid._k_v, self.sphere._R_0,
        )

        self._x_hat.is_computed_by(
            self.scatterer._k_f, self.sphere._R_0,
        )
        self._x_hat_v.is_computed_by(
            self.scatterer._k_v, self.sphere._R_0,
        )

        # determinants
        self._det_M_n.is_computed_by(
            #  self._n,
            self._m_11, self._m_12, self._m_13, self._m_14,
            self._m_21, self._m_22, self._m_23, self._m_24,
            self._m_31, self._m_32, self._m_33, self._m_34,
            self._m_41, self._m_42, self._m_43, self._m_44,
        )

        self._det_M_1_n.is_computed_by(
            #  self._n,
            self._m_21, self._m_22, self._m_23, self._m_24,
            self._m_31, self._m_32, self._m_33, self._m_34,
            self._m_41, self._m_42, self._m_43, self._m_44,
            self._n_1, self._n_2, self._n_3, self._n_4,
        )

        self._det_M_2_n.is_computed_by(
            #  self._n,
            self._m_11, self._m_12, self._m_13, self._m_14,
            self._n_1, self._n_2, self._n_3, self._n_4,
            self._m_31, self._m_32, self._m_33, self._m_34,
            self._m_41, self._m_42, self._m_43, self._m_44,
        )

        self._det_M_3_n.is_computed_by(
            #  self._n,
            self._m_11, self._m_12, self._m_13, self._m_14,
            self._m_21, self._m_22, self._m_23, self._m_24,
            self._n_1, self._n_2, self._n_3, self._n_4,
            self._m_41, self._m_42, self._m_43, self._m_44,
        )
        self._det_M_4_n.is_computed_by(
            #  self._n,
            self._m_11, self._m_12, self._m_13, self._m_14,
            self._m_21, self._m_22, self._m_23, self._m_24,
            self._m_31, self._m_32, self._m_33, self._m_34,
            self._n_1, self._n_2, self._n_3, self._n_4,
        )

        # RHS vector
        self._n_1.is_computed_by(self._x)
        self._n_2.is_computed_by(self._x)
        self._n_3.is_computed_by(
            self._x, self.frequency._omega,
            self.fluid._eta_f,
            self.fluid._zeta_f,
        )
        self._n_4.is_computed_by(self._x, self.fluid._eta_f)

        # matrix entries m_1j
        self._m_11.is_computed_by(self._x)
        self._m_12.is_computed_by(self._x_v)
        self._m_13.is_computed_by(self._x_hat)
        self._m_14.is_computed_by(self._x_hat_v)

        # matrix entries m_2j
        self._m_21.is_computed_by(self._x)
        self._m_22.is_computed_by(self._x_v)
        self._m_23.is_computed_by(self._x_hat)
        self._m_24.is_computed_by(self._x_hat_v)

        # matrix entries m_3j
        self._m_31.is_computed_by(
            self._x, self.fluid._eta_f, self.fluid._zeta_f, self.fluid._rho_f,
            self.fluid._c_f, self.frequency._omega,
        )
        self._m_32.is_computed_by(self._x_v, self.fluid._eta_f)
        self._m_33.is_computed_by(
            self._x_hat, self.scatterer._eta_f, self.scatterer._zeta_f,
            self.scatterer._rho_f, self.scatterer._c_f, self.frequency._omega,
        )
        self._m_34.is_computed_by(self._x_hat_v, self.scatterer._eta_f)

        # matrix entries m_3j
        self._m_41.is_computed_by(self._x, self.fluid._eta_f)
        self._m_42.is_computed_by(self._x_v, self.fluid._eta_f)
        self._m_43.is_computed_by(self._x_hat, self.scatterer._eta_f)
        self._m_44.is_computed_by(self._x_hat_v, self.scatterer._eta_f)

    def __repr__(self):
        return (
            f'Doinikov1994Compressible.ScatteringFiels(frequency={self.f}, '
            f'radius={self.R_0}, '
            f'rho_s={self.rho_s}, c_s={self.c_s}, '
            f'eta_s={self.eta_s}, zeta_s={self.zeta_s}, '
            f'rho_f={self.rho_f}, c_f={self.c_f}, '
            f'eta_f={self.eta_f}, zeta_f={self.zeta_f}, '
            f'p_0={self.p_0}, position={self.position}, {self.wave_type}, '
            f'N_max={self.N_max})'
        )
    # -----------------------------------------------------
    # Determinants of matrices
    # -----------------------------------------------------

    def det_M_n(self, n: int) -> complex:
        """Determinant of M matrix of order n

        :param n: order
        """
        if n < len(self._det_M_n.value):
            return self._det_M_n.value[n]
        else:
            self._compute_det_M_n(n)
            return self._det_M_n.value[n]

    def _compute_det_M_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._det_M_n.value)
        for n in full_range(N_old, N):
            self._det_M_n.value.append(self.det_M(n))

    def det_M_1_n(self, n: int) -> complex:
        """Determinant of M_1 matrix of order n

        :param n: order
        """
        if n < len(self._det_M_1_n.value):
            return self._det_M_1_n.value[n]
        else:
            self._compute_det_M_1_n(n)
            return self._det_M_1_n.value[n]

    def _compute_det_M_1_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._det_M_1_n.value)
        for n in full_range(N_old, N):
            self._det_M_1_n.value.append(self.det_M_1(n))

    def det_M_2_n(self, n: int) -> complex:
        """Determinant of M_2 matrix of order n

        :param n: order
        """
        if n < len(self._det_M_2_n.value):
            return self._det_M_2_n.value[n]
        else:
            self._compute_det_M_2_n(n)
            return self._det_M_2_n.value[n]

    def _compute_det_M_2_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._det_M_2_n.value)
        for n in full_range(N_old, N):
            self._det_M_2_n.value.append(self.det_M_2(n))

    def det_M_3_n(self, n: int) -> complex:
        """Determinant of M_3 matrix of order n

        :param n: order
        """
        if n < len(self._det_M_3_n.value):
            return self._det_M_3_n.value[n]
        else:
            self._compute_det_M_3_n(n)
            return self._det_M_3_n.value[n]

    def _compute_det_M_3_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._det_M_3_n.value)
        for n in full_range(N_old, N):
            self._det_M_3_n.value.append(self.det_M_3(n))

    def det_M_4_n(self, n: int) -> complex:
        """Determinant of M_4 matrix of order n

        :param n: order
        """
        if n < len(self._det_M_4_n.value):
            return self._det_M_4_n.value[n]
        else:
            self._compute_det_M_4_n(n)
            return self._det_M_4_n.value[n]

    def _compute_det_M_4_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._det_M_4_n.value)
        for n in full_range(N_old, N):
            self._det_M_4_n.value.append(self.det_M_4(n))

    # -----------------------------------------------------
    # Scattering coefficients Fluid
    # -----------------------------------------------------

    def alpha_n(self, n: int) -> complex:
        """ coefficient :math:`\\alpha_n` (3.13) and (3.20)

        :param n: order
        """
        if n < len(self._alpha_n.value):
            return self._alpha_n.value[n]
        else:
            self._compute_alpha_n(n)
            return self._alpha_n.value[n]

    def _compute_alpha_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._alpha_n.value)
        for n in full_range(N_old, N):
            if n == 0:
                tmp1 = 4 * (self.eta_f - self.eta_s) * self.x * self.x_hat
                tmp1 *= Sp.besselj(1, self.x_hat)

                tmp2 = 1j * self.omega * self.R_0**2

                # numerator
                num = self.rho_f * self.x_hat
                num *= Sp.besselj(0, self.x) * Sp.besselj(1, self.x_hat)

                num -= self.rho_s * self.x * (
                    Sp.besselj(1, self.x) *
                    Sp.besselj(0, self.x_hat)
                )

                num *= tmp2

                num -= tmp1 * Sp.besselj(1, self.x)

                # denominator
                den = self.rho_s * self.x
                den *= Sp.hankelh1(1, self.x) * Sp.besselj(0, self.x_hat)

                den -= self.rho_f * self.x_hat * (
                    Sp.hankelh1(0, self.x) *
                    Sp.besselj(1, self.x_hat)
                )

                den *= tmp2

                den += tmp1 * Sp.hankelh1(1, self.x)

                new_value = num / den * self.A_in(0)
            else:
                new_value = self.det_M_1_n(n) / self.det_M_n(n)

            self._alpha_n.value.append(new_value)

    def beta_n(self, n: int) -> complex:
        """ coefficient :math:`\\beta_n` (3.13) and (3.20)

        :param n: order
        """
        if n < len(self._beta_n.value):
            return self._beta_n.value[n]
        else:
            self._compute_beta_n(n)
            return self._beta_n.value[n]

    def _compute_beta_n(self, N: int) -> None:
        # computation according to (3.13), (3.15) and (3.21)

        N_old = len(self._beta_n.value)
        for n in full_range(N_old, N):
            if n == 0:
                new_value = 0.0
            else:
                new_value = self.det_M_2_n(n) / self.det_M_n(n)

            self._beta_n.value.append(new_value)

    # -----------------------------------------------------
    # Scattering coefficients sphere
    # -----------------------------------------------------

    def alpha_hat_n(self, n: int) -> complex:
        """ coefficient :math:`\\alpha_hat_n` (3.13) and (3.20)

        :param n: order
        """
        if n < len(self._alpha_hat_n.value):
            return self._alpha_hat_n.value[n]
        else:
            self._compute_alpha_hat_n(n)
            return self._alpha_hat_n.value[n]

    def _compute_alpha_hat_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._alpha_hat_n.value)
        for n in full_range(N_old, N):
            if n == 0:
                tmp1 = 4 * (self.eta_f - self.eta_s) * self.x * self.x_hat
                tmp2 = 1j * self.omega * self.R_0**2

                # numerator
                num = Sp.besselj(0, self.x) * Sp.hankelh1(1, self.x)

                num -= Sp.besselj(1, self.x) * Sp.hankelh1(0, self.x)

                num *= tmp2 * self.x * self.rho_f

                # denominator
                den = self.rho_s * self.x
                den *= Sp.hankelh1(1, self.x) * Sp.besselj(0, self.x_hat)

                den -= self.rho_f * self.x_hat * (
                    Sp.hankelh1(0, self.x) *
                    Sp.besselj(1, self.x_hat)
                )

                den *= tmp2

                den += tmp1 * (
                    Sp.hankelh1(1, self.x) *
                    Sp.besselj(1, self.x_hat)
                )

                new_value = num / den * self.A_in(0)
            else:
                new_value = self.det_M_3_n(n) / self.det_M_n(n)

            self._alpha_hat_n.value.append(new_value)

    def beta_hat_n(self, n: int) -> complex:
        """ coefficient :math:`\\beta_hat_n` (3.13) and (3.20)

        :param n: order
        """
        if n < len(self._beta_hat_n.value):
            return self._beta_hat_n.value[n]
        else:
            self._compute_beta_hat_n(n)
            return self._beta_hat_n.value[n]

    def _compute_beta_hat_n(self, N: int) -> None:
        # computation according to (3.13), (3.15) and (3.21)

        N_old = len(self._beta_hat_n.value)
        for n in full_range(N_old, N):
            if n == 0:
                new_value = 0.0
            else:
                new_value = self.det_M_4_n(n) / self.det_M_n(n)

            self._beta_hat_n.value.append(new_value)

    # -----------------------------------------------------
    # Scattering coefficients sphere
    # -----------------------------------------------------

    @staticmethod
    def _reset_list() -> list:
        return []

    @property
    def x(self) -> complex:
        """Product of :attr:`~.k_f` and :attr:`~.R_0`
        """
        return self._x.value

    def _compute_x(self) -> complex:
        return self.k_f * self.R_0

    @property
    def x_v(self) -> complex:
        """Product of :attr:`~.k_v` and :attr:`~.R_0`
        """
        return self._x_v.value

    def _compute_x_v(self) -> complex:
        return self.k_v * self.R_0

    @property
    def x_hat(self) -> complex:
        """Product of :attr:`~.k_f` and :attr:`~.R_0`
        """
        return self._x_hat.value

    def _compute_x_hat(self) -> complex:
        return self.k_s * self.R_0

    @property
    def x_hat_v(self) -> complex:
        """Product of :attr:`~.k_v` and :attr:`~.R_0`
        """
        return self._x_hat_v.value

    def _compute_x_hat_v(self) -> complex:
        return self.k_vs * self.R_0

    # -----------------------------------------------------
    # Methods
    # -----------------------------------------------------

    def V_r_sc(self, n: int, r: float) -> complex:
        """Radial scattering field velocity term of mode `n`
        without Legendre coefficients

        Returns radial scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        out = self.k_f * self.alpha_n(n) * Sp.d1_hankelh1(n, r * self.k_f)
        out -= n * (n + 1) / r * self.beta_n(n) * Sp.hankelh1(n, r * self.k_v)
        #  out *= self.A_in(n)
        return out

    def V_theta_sc(self, n: int, r: float) -> complex:
        """ Tangential scattering field velocity term of mode n
        without Legendre coefficients

        Returns tangential scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """

        arg = self.k_v * r
        out = Sp.hankelh1(n, arg)
        out += arg * Sp.d1_hankelh1(n, arg)
        out *= -self.beta_n(n)

        out += self.alpha_n(n) * Sp.hankelh1(n, self.k_f * r)

        #  out *= self.A_in(n)

        return out / r

    def radial_particle_velocity(
        self,
        r: Union[float, np.ndarray, list[float]],
        theta: Union[float, np.ndarray, list[float]],
        t: Union[float, np.ndarray, list[float]],
        mode: Optional[int] = None,
    ) -> float:
        """Particle velocity in radial direction

        Returns the value of the particle velocity
        in radial direction in [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific modenumber of interest; if `None` then all
                     modes until :attr:`.N_max`
        """
        r, theta, t = InputHandler.handle_input(
            r, theta, t, self.R_0,
            inside_sphere=True,
        )

        def radial_velocity(n: int, r: float) -> complex:
            out = self.k_s * self.alpha_hat_n(n)
            out *= Sp.d1_besselj(n, r * self.k_s)

            arg = r * self.k_vs
            out -= n * (n + 1) / r * self.beta_hat_n(n) * Sp.besselj(n, arg)

            return out

        out = self.radial_mode_superposition(
            radial_velocity, r, theta, t, mode,
        )

        return out

    def tangential_particle_velocity(
            self,
            r: Union[float, np.ndarray, list[float]],
            theta: Union[float, np.ndarray, list[float]],
            t: Union[float, np.ndarray, list[float]],
            mode: Optional[int] = None,
    ) -> float:
        """Particle velocity in tangential direction

        Returns the value of the particle velocity
        in tangential direction in [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: specific modenumber of interest; if `None` then all
                     modes until :attr:`.N_max`
        """
        r, theta, t = InputHandler.handle_input(
            r, theta, t, self.R_0,
            inside_sphere=True,
        )

        def tangential_velocity(n: int, r: float) -> complex:
            out = Sp.besselj(n, self.k_vs * r)
            out += r * self.k_vs * Sp.d1_besselj(n, self.k_vs * r)
            out *= -self.beta_hat_n(n)

            out += self.alpha_hat_n(n) * Sp.besselj(n, self.k_s * r)

            return out / r

        out = self.tangential_mode_superposition(
            tangential_velocity, r, theta, t, mode,
        )

        return out


if __name__ == '__main__':
    pass



import unittest

from osaft.core.frequency import Frequency
from osaft.core.functions import sqrt
from osaft.core.solids import ElasticSolid, RigidSolid
from osaft.tests.basetest import BaseTest


class TestRigidSolid(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.cls = RigidSolid(self.f, self.rho_s)
        self.composite_cls = RigidSolid(self.frequency, self.rho_s)
        self.list_cls = [self.cls, self.composite_cls]

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_properties(self):
        self.do_testing(lambda: self.rho_s, lambda: self.cls.rho_s)
        self.do_testing(
            lambda: self.cls.rho_s,
            lambda: self.composite_cls.rho_s,
        )

        self.do_testing(
            lambda: self.f,
            lambda: self.cls.f,
        )
        self.do_testing(
            lambda: self.cls.f,
            lambda: self.composite_cls.f,
        )


class TestElasticSolid(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.frequency = Frequency(self.f)
        self.cls = ElasticSolid(self.f, self.E_s, self.nu_s, self.rho_s)
        self.composite_cls = ElasticSolid(
            self.frequency,
            self.E_s, self.nu_s, self.rho_s,
        )
        self.list_cls = [self.cls, self.composite_cls]
    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def compute_B_s(self) -> float:
        return self.E_s / 3 / (1 - 2 * self.nu_s)

    def compute_kappa_s(self) -> float:
        return 3 * (1 - 2 * self.nu_s) / self.E_s

    def compute_c_l(self) -> float:
        return sqrt(
            self.E_s * (1 - self.nu_s)
            / (self.rho_s * (1 + self.nu_s) * (1 - 2 * self.nu_s)),
        )

    def compute_c_t(self) -> float:
        return sqrt(self.E_s / (2 * (1 + self.nu_s) * self.rho_s))

    def compute_k_l(self) -> float:
        omega = self.frequency.omega
        c_l = self.compute_c_l()
        return omega / c_l

    def compute_k_t(self) -> float:
        omega = self.frequency.omega
        c_t = self.compute_c_t()
        return omega / c_t

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_c_l(self) -> None:
        self.do_testing(
            func_1=self.compute_c_l,
            func_2=lambda: self.cls.c_l,
        )
        self.do_testing(
            func_1=self.compute_c_l,
            func_2=lambda: self.composite_cls.c_l,
        )

    def test_c_t(self) -> None:
        self.do_testing(
            func_1=self.compute_c_t,
            func_2=lambda: self.cls.c_t,
        )
        self.do_testing(
            func_1=self.compute_c_t,
            func_2=lambda: self.composite_cls.c_t,
        )

    def test_k_l(self) -> None:
        self.do_testing(
            func_1=self.compute_k_l,
            func_2=lambda: self.cls.k_l,
        )
        self.do_testing(
            func_1=self.compute_k_l,
            func_2=lambda: self.composite_cls.k_l,
        )

    def test_k_t(self) -> None:
        self.do_testing(
            func_1=self.compute_k_t,
            func_2=lambda: self.cls.k_t,
        )
        self.do_testing(
            func_1=self.compute_k_t,
            func_2=lambda: self.composite_cls.k_t,
        )

    def test_kappa_s(self) -> None:
        self.do_testing(
            func_1=self.compute_kappa_s,
            func_2=lambda: self.cls.kappa_s,
        )
        self.do_testing(
            func_1=self.compute_kappa_s,
            func_2=lambda: self.composite_cls.kappa_s,
        )

    def test_B_s(self) -> None:
        self.do_testing(
            func_1=self.compute_B_s,
            func_2=lambda: self.cls.B_s,
        )
        self.do_testing(
            func_1=self.compute_B_s,
            func_2=lambda: self.composite_cls.B_s,
        )


if __name__ == '__main__':
    unittest.main()



import time
import unittest
from abc import abstractmethod
from collections.abc import Callable, Sequence
from numbers import Number
from typing import Any, Optional, Union

import numpy as np

from osaft import WaveTypes
from osaft.core.frequency import Frequency
from osaft.solutions.base_solution import BaseSolution


class BaseChangingVariable:
    """Type used as attribute of a `TestCase` that is changed automatically.

        To test the observer pattern testing, attributes of classes are
        changed to ensure proper linking. If a test class has an attribute of
        type :class:`BaseChangingVariable` then the attribute of the tested
        class with the name `name` is initially set to `value`. During the
        testing the attribute is changed to a new value using the method
        `change`.

        :param name: name of the attribute in the class to be tested
        :param value: initial value of the attribute
        :param seed: seed for RNG to make testing reproducible
    """

    def __init__(
        self, name: str, value: Any,
        seed: int,
    ) -> None:
        """Constructor method
        """
        self.name = name
        self.value = value

        self.rng = np.random.default_rng(seed)

    @abstractmethod
    def change(self) -> None:
        """Randomly change `value`
        """
        pass


class ChangingFromList(BaseChangingVariable):
    """Child of :class:`BaseChangingVariable`.

    The method `change` changes `value` by choosing an item from
    :attr:`list_of_values`.

    :param name: name of the attribute in the class to be tested
    :param value: initial value of the attribute
    :param list_of_values: list of values to choose from
    :param seed: seed for RNG to make testing reproducible
    """
    _longest_list = 0

    def __init__(
        self, name: str, value: Any,
        list_of_values: Sequence[Any],
        seed: int,
    ) -> None:

        super().__init__(name, value, seed)

        self.list_of_values = list_of_values

    def _set_longest_list(self) -> None:
        if len(self.list_of_values) > self._longest_list:
            ChangingFromList._longest_list = len(self.list_of_values)

    def _reset_non_tested_values(self) -> None:
        # [*list] syntax is needed for not passing a reference but creating an
        # actual new list
        self._non_tested_values = [*self.list_of_values]

    @property
    def list_of_values(self) -> list[any]:
        return self._list_of_values

    @list_of_values.setter
    def list_of_values(self, lst: list[any]) -> None:
        self._list_of_values = lst
        self._set_longest_list()
        self._reset_non_tested_values()

    @property
    def longest_list(self) -> int:
        return ChangingFromList._longest_list

    def change(self) -> None:
        """Randomly change `value`
        """
        if len(self._non_tested_values) < 1:
            self._reset_non_tested_values()

        self.value = self.rng.choice(self._non_tested_values)
        self._non_tested_values.remove(self.value)


class ChangingBool(BaseChangingVariable):
    """Child of :class:`BaseChangingVariable`

    Child of :class:`BaseChangingVariable` for the case when value is of type
    `bool`.

    :param name: name of the attribute in the class to be tested
    :param value: initial value of the attribute
    :param seed: seed for RNG to make testing reproducible
    """

    def __init__(self, name: str, value: bool, seed: int) -> None:
        """Constructor method
        """
        super().__init__(name, value, seed)

    @abstractmethod
    def change(self) -> None:
        """Randomly change `value`
        """
        self.value = self.rng.random() > 0.5


class ChangingNumber(BaseChangingVariable):
    """Child of :class:`BaseChangingVariable`

    Child of :class:`BaseChangingVariable` for the case when value is a number.

    :param name: name of the attribute in the class to be tested
    :param value: initial value of the attribute
    :param seed: seed for RNG to make testing reproducible
    :param low: lowest value `value` is changed to
    :param high: highest value `value` is changed to
    """

    def __init__(
        self, name: str, value: Union[None, float, int], seed: int,
        low: Union[None, float, int], high: Union[float, int],
    ) -> None:
        """Constructor method
        """
        super().__init__(name, value, seed)
        self.test_inputs(value, low, high)
        self._low = low
        self._high = high

    @staticmethod
    def test_inputs(
        value: Union[None, int, float],
        low: Union[None, int, float],
        high: Union[None, int, float],
    ):
        if low is not None and high is not None:
            assert low < value < high, 'condition low < value < high not met'
        elif low is not None:
            assert low < value, 'condition low < value < high is not met'
        elif high is not None:
            assert value < high, 'condition low < value < high is not met'

    @property
    def low(self) -> Union[None, int, float]:
        return self._low

    @low.setter
    def low(self, value: Union[int, float]) -> None:
        if self.high is not None and value > self.high:
            raise ValueError(
                f'high (= {value}) needs to be larger than low '
                f'(= {self.low})',
            )
        elif value > self.value:
            self._low = value
            self.value = value
            self.change()
        else:
            self._low = value

    @property
    def high(self) -> Union[float, int]:
        return self._high

    @high.setter
    def high(self, value: Union[int, float]) -> None:
        if self.low is not None and value < self.low:
            raise ValueError(
                f'high (= {value}) needs to be larger than low '
                f'(= {self.low})',
            )
        elif value < self.value:
            self._high = value
            self.value = value
            self.change()
        else:
            self._high = value

    @abstractmethod
    def change(self) -> None:
        """Randomly change `value`
        """
        pass


class ChangingFloat(ChangingNumber):
    """Child of :class:`BaseChangingVariable`

    Child of :class:`BaseChangingVariable` for the case when value is a
    `float`.

    :param name: name of the attribute in the class to be tested
    :param value: initial value of the attribute
    :param seed: seed for RNG to make testing reproducible
    :param low: lowest value `value` is changed to
    :param high: highest value `value` is changed to
    """

    def __init__(
        self, name: str, value: float, seed: int,
        low: Optional[float] = None,
        high: Optional[float] = None,
    ) -> None:
        """Constructor method
        """
        super().__init__(name, value, seed, low, high)

    def change(self) -> None:
        """Randomly change `value`
        """
        if self.low is not None and self.high is not None:
            self.value = (
                self.low + (self.high - self.low) *
                self.rng.random()
            )
        elif self.low is not None:
            change = self.value * (self.rng.random() - 0.5)
            if self.value + change > self.low:
                self.value += change
            else:
                self.value -= change
        elif self.high is not None:
            change = self.value * (self.rng.random() - 0.5)
            if self.value + change < self.high:
                self.value += change
            else:
                self.value -= change
        else:
            self.value += (self.rng.random() - 0.5) * self.value


class ChangingInt(ChangingNumber):
    """Child of :class:`BaseChangingVariable`

    Child of :class:`BaseChangingVariable` for the case when value is an
    `int`. Default values for `low` and `high` are 0 and 100 respectively. If
    `value` is outside this range the parameters have to be set accordingly.

    :param name: name of the attribute in the class to be tested
    :param value: initial value of the attribute
    :param seed: seed for RNG to make testing reproducible
    :param low: lowest value `value` is changed to
    :param high: highest value `value` is changed to
    """

    def __init__(
        self, name: str, value: int, seed: int, low: int = 0,
        high: int = 100,
    ):
        """Constructor method
        """
        super().__init__(name, value, seed, low, high)

    def change(self) -> None:
        """Randomly change `value`
        """
        self.value = self.rng.integers(low=self.low, high=self.high)


class NumericTestCase(unittest.TestCase):

    def assertAlmostEqual(
        self,
        val1: Number,
        val2: Number,
        threshold: Optional[float] = 1e-12,
        print_results: Optional[bool] = False,
        user_msg: Optional[str] = None,
    ) -> None:
        """ Asserting the relative difference of two values when num. rounding
        errors can occur

        Assertion is handled by :meth:`BaseTest._assertAlmostEqual()`

        If the values are complex, then the real and imaginary part will be
        tested separately. Also, if one of the test fails, automatically, a
        nice formatted output is generated as error message.

        :param val1: first value to be compared
        :param val2: second value to be compared
        :param threshold: threshold of relative difference, defaults to 1e-12
        :param print_results: if True the two values are printed to the console
            for easier debugging, defaults to False
        :param user_msg: custom message that is printed if the assertion fails.
        """
        if print_results:
            msg = f'\nval1     : {val1:+.8e}\n'
            msg += f'val2     : {val2:+.8e}\n'
            if not val1 == 0 and not val2 == 0:
                msg += f'rel diff.: {abs((val1 - val2) / val1):+.8e}'
            print(msg)

        self._assertAlmostEqual(val1.real, val2.real, threshold)
        if isinstance(val1, complex) and isinstance(val2, complex):
            self._assertAlmostEqual(val1.imag, val2.imag, threshold, user_msg)

    def _assertAlmostEqual(
        self, val1: float, val2: float,
        threshold: float, user_msg: Optional[str] = None,
    ) -> None:

        msg = '\nAssertion failed\n'
        if user_msg:
            msg += f'\n{user_msg}\n'

        if val1 == 0 or val2 == 0:
            msg += 'One of the values is Zero; the other not\n'
            msg += f'val1     : {val1:+.8e}\n'
            msg += f'val2     : {val2:+.8e}\n'
            test = abs(val1) < threshold and abs(val2) < threshold
            self.assertTrue(expr=test, msg=msg)
        else:
            diff = np.abs(val1 - val2)
            if np.abs(val1) > np.abs(val2):
                test = np.abs(diff / val2) < threshold
            else:
                test = np.abs(diff / val1) < threshold

            msg += 'The rel. diff. is bigger than the threshold\n'
            msg += f'val1     : {val1:+.8e}\n'
            msg += f'val2     : {val2:+.8e}\n'
            msg += f'rel diff.: {abs(diff/val1):+.8e}\n'
            msg += f'threshold: {threshold:+.8e}'

            self.assertTrue(expr=test, msg=msg)


class BaseTest(NumericTestCase):

    def setUp(self) -> None:

        # Create
        seed = np.random.randint(1, 10000)
        self._seed = seed

        # set of all passive attributes that have a value of type float
        # that appear in classes in this module

        # Geometry
        self._R_0 = ChangingFloat('R_0', 1e-6, seed, low=0)
        # Frequency
        self._f = ChangingFloat('f', 1e6, seed, low=0)
        # Inviscid Fluid
        self._c_f = ChangingFloat('c_f', 1.5e3, seed, low=0)
        self._rho_f = ChangingFloat('rho_f', 1e3, seed, low=0)
        # Compressible Particle (Inviscid Fluid)
        self._c_s = ChangingFloat('c_s', 2.5e3, seed, low=0)
        # Compressible Particle (Viscous Fluid)
        self._eta_s = ChangingFloat('eta_s', 1e-3, seed, low=0)
        self._zeta_s = ChangingFloat('zeta_s', 1e-3, seed, low=0)
        # Viscous Fluid
        self._eta_f = ChangingFloat('eta_f', 1e-2, seed, low=0)
        self._zeta_f = ChangingFloat('zeta_f', 1e-2, seed, low=0)
        # Viscoelastic Fluid
        self._eta_p = ChangingFloat('eta_p', 1e-3, seed, low=0)
        self._zeta_p = ChangingFloat('zeta_p', 1e-3, seed, low=0)
        self._lambda_M = ChangingFloat('lambda_M', 1e-3, seed, low=0)
        # Rigid Solid
        self._rho_s = ChangingFloat('rho_s', 1.5e3, seed, low=0)
        # Elastic solid
        self._E_s = ChangingFloat('E_s', 75e6, seed, low=0)
        self._nu_s = ChangingFloat('nu_s', 0.3, seed, low=0, high=0.49)
        # Background Field
        self._p_0 = ChangingFloat('p_0', 1e5, seed, low=0)
        self._position = ChangingFloat(
            'position', np.pi / 4, seed, low=0, high=2 * np.pi,
        )
        self._wave_type = ChangingFromList(
            'wave_type', WaveTypes.STANDING,
            WaveTypes, seed,
        )

        self.test_variables = {
            # Geometry
            self._R_0,
            # Frequency
            self._f,
            # Inviscid Fluid
            self._c_f,
            self._rho_f,
            # Compressible Particle (Inviscid Fluid)
            self._c_s,
            self._rho_s,
            self._eta_s,
            self._zeta_s,
            # Viscous Fluid
            self._eta_f,
            self._zeta_f,
            # Viscoelastic Fluid
            self._eta_p,
            self._zeta_p,
            self._lambda_M,
            # Elastic solid
            self._E_s,
            self._nu_s,
            # Background Field
            self._p_0,
            self._position,
            self._wave_type,
        }

        # Frequency Class
        self.frequency = Frequency(self.f)

        # List of classes to be tests
        self.list_cls = []

        # timing
        self._time = time.time()

        # Runs
        self._n_runs = self._wave_type.longest_list

    @property
    def n_runs(self) -> int:
        return self._n_runs

    @n_runs.setter
    def n_runs(self, val) -> None:
        if val < self._wave_type.longest_list:
            raise ValueError(
                'n_runs must be greater than the longest '
                'list of ChangingFromList._longest_list',
            )
        else:
            self._n_runs = val

    def tearDown(self) -> None:
        run_time = time.time() - self._time
        print(f'\n\trun_time: {run_time*1e3:9.4f}ms', end=' ', flush=True)
        if run_time > 0.1:
            print('>>>>TEST TOOK LONGER THAN 100ms<<<<', end=' ', flush=True)

    def test_name_attribute(self) -> None:
        if hasattr(self, 'cls'):
            if isinstance(self.cls, BaseSolution):
                self.assertTrue(self.cls.name != '')
    # -------------------------------------------------------------------------
    # Generic test methods for properties and methods
    # -------------------------------------------------------------------------

    def _test_properties(
            self,
            list_of_properties: list[str],
            threshold: float = 1e-12,
    ) -> None:
        """Tests the properties inside `list_of_properties` of self & self.cls

        If a test fails the name of the failing property is printed.

        :param list_of_properties: list of properties to be tested
        :param threshold: comparing threshold for the whole list
        """

        for name in list_of_properties:
            with self.subTest(property=name):
                self.do_testing(
                    func_1=lambda: getattr(self, name),
                    func_2=lambda: getattr(self.cls, name),
                    threshold=threshold,
                )

    def _test_methods_n(
            self,
            dict_of_methods: dict[str, dict[tuple[Any]]],
            n_end: int = 5,
            threshold: float = 1e-12,
    ) -> None:
        """Tests the methods inside `dict_of_methods` of self & self.cls

        All methods have as first argument `n`. The key of the dict is the name
        of the methods. The `value` of the dict are the optional additional
        arguments.  If `value` == None, then it is a method solely depending on
        `n`.

        If a test fails the name of the failing method is printed.

        :param dict_of_methods: dict of methods and additional arguments to be
            tested
        :param n_end: n in [0, 1, ..., n_end]; tested in reversed order
        :param threshold: comparing threshold for the whole dict
        """

        for method, args in dict_of_methods.items():
            func_1 = getattr(self, method)
            func_2 = getattr(self.cls, method)
            for n in reversed(np.arange(n_end + 1)):

                if args is not None:
                    if isinstance(args, tuple):
                        arg = (n, *args)
                    else:
                        arg = (n, args)
                else:
                    arg = (n,)

                with self.subTest(method=method, arguments=arg):
                    self.do_testing(
                        func_1=func_1, args_1=arg,
                        func_2=func_2, args_2=arg,
                        threshold=threshold,
                    )

    # -------------------------------------------------------------------------
    # Getters & Setters
    # -------------------------------------------------------------------------

    @property
    def R_0(self) -> float:
        return self._R_0.value

    @R_0.setter
    def R_0(self, value: float) -> None:
        self._R_0.value = value

    @property
    def f(self) -> float:
        return self._f.value

    @f.setter
    def f(self, value: float) -> None:
        self._f.value = value

    @property
    def c_f(self) -> float:
        return self._c_f.value

    @c_f.setter
    def c_f(self, value: float) -> None:
        self._c_f.value = value

    @property
    def rho_f(self) -> float:
        return self._rho_f.value

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self._rho_f.value = value

    @property
    def c_s(self) -> float:
        return self._c_s.value

    @c_s.setter
    def c_s(self, value: float) -> None:
        self._c_s.value = value

    @property
    def eta_s(self) -> float:
        return self._eta_s.value

    @eta_s.setter
    def eta_s(self, value: float) -> None:
        self._eta_s.value = value

    @property
    def zeta_s(self) -> float:
        return self._zeta_s.value

    @zeta_s.setter
    def zeta_s(self, value: float) -> None:
        self._eta_s.value = value

    @property
    def rho_s(self) -> float:
        return self._rho_s.value

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self._rho_s.value = value

    @property
    def eta_f(self) -> float:
        return self._eta_f.value

    @eta_f.setter
    def eta_f(self, value: float) -> None:
        self._eta_f.value = value

    @property
    def zeta_f(self) -> float:
        return self._zeta_f.value

    @zeta_f.setter
    def zeta_f(self, value: float) -> None:
        self._zeta_f.value = value

    @property
    def eta_p(self) -> float:
        return self._eta_p.value

    @eta_p.setter
    def eta_p(self, value: float) -> None:
        self._eta_p.value = value

    @property
    def zeta_p(self) -> float:
        return self._zeta_p.value

    @zeta_p.setter
    def zeta_p(self, value: float) -> None:
        self._zeta_p.value = value

    @property
    def lambda_M(self) -> float:
        return self._lambda_M.value

    @lambda_M.setter
    def lambda_M(self, value: float) -> None:
        self._lambda_M.value = value

    @property
    def E_s(self) -> float:
        return self._E_s.value

    @E_s.setter
    def E_s(self, value: float) -> None:
        self._E_s.value = value

    @property
    def nu_s(self) -> float:
        return self._nu_s.value

    @nu_s.setter
    def nu_s(self, value: float) -> None:
        self._nu_s.value = value

    @property
    def p_0(self) -> float:
        return self._p_0.value

    @p_0.setter
    def p_0(self, value: float) -> None:
        self._p_0.value = value

    @property
    def position(self) -> float:
        return self._position.value

    @position.setter
    def position(self, value: float) -> None:
        self._position.value = value

    @property
    def wave_type(self) -> WaveTypes:
        return self._wave_type.value

    @wave_type.setter
    def wave_type(self, value: WaveTypes) -> None:
        self._wave_type.value = value

    # -----------------------------------------------------------------------------
    # Tests
    # -----------------------------------------------------------------------------

    def do_testing(
        self, func_1: Callable, func_2: Callable,
        args_1: Optional[Any] = None,
        args_2: Optional[Any] = None,
        threshold: float = 1e-10,
        zero: Optional[float] = None,
    ) -> None:
        """Compares the outputs of `func_1` against `func_2`.

        The variables in
        :attr:`BaseTest.test_variables` are changed and assigned one
        after the other to test all dependencies. This procedure is repeated
        :attr:`BaseTest.n_runs` times. Arguments to the functions can be
        passed with `args_1` and `args_2` to the respective functions.
        Testing is handled by :meth:`BaseTest._test_variables()` and changing
        of the values :meth:`BaseTest.change_and_assign_single_variable`

        In order to test properties of a class the following syntax is used:

        .. highlight:: python

        self.do_testing(lambda: self.cls.some_property, self.some_function)

        :param func_1: first function
        :param func_2: second function
        :param args_1: arguments for first function
        :param args_2: arguments for second functions
        :param threshold: threshold for failing the test, defaults to 1e-12
        :param zero: value below which test values are assumed to be zero
        """
        self.assign_parameters()
        self._test_variables(
            func_1, args_1, func_2, args_2,
            threshold, zero,
        )
        for _ in np.arange(self.n_runs):
            for var in self.test_variables:
                if self._check_if_has_attr(var.name):
                    with self.subTest(msg=f'Changing {var.name}'):
                        self.change_and_assign_single_variable(var)
                        self._test_variables(
                            func_1, args_1, func_2, args_2,
                            threshold, zero,
                        )

    def _check_if_has_attr(self, name: str):
        """Check if any class in list_cls has the attribute `name`

        param name: name of the attribute
        """
        list_hasattr = [hasattr(cls, name) for cls in self.list_cls]
        return np.any(list_hasattr)

    def _test_variables(
        self,
        func_1: Callable,
        args_1: tuple,
        func_2: Callable,
        args_2: tuple,
        threshold: Optional[float] = 1e-12,
        zero: Optional[float] = None,
    ) -> None:
        """Checks if func_1(args) is almost equal to func_2 args

        Checks if func_1(args) is almost equal to func_2 args with tolerance
        `threshold`. Values below `zero` are assumed to be 0.

        :param func_1: first function
        :param args_1: arguments for the first function
        :param func_2: second function
        :param args_2: arguments for the second function
        :param threshold: tolerance
        "param zero: value below which test values are assumed to be zero
        """
        # check type of parameters1
        first = self._get_value(func_1, args_1)
        second = self._get_value(func_2, args_2)

        if zero is not None:
            first = 0 if abs(first) < zero else first
            second = 0 if abs(second) < zero else second

        self.assertAlmostEqual(first, second, threshold=threshold)

    @staticmethod
    def _get_value(func: Callable, args: tuple):
        """Get the value of the function `func` with arguments `arg`

        :param func: function
        :param args: arguments
        """
        if args is None:
            out = func()
        elif isinstance(args, tuple):
            out = func(*args)
        else:
            out = func(args)
        return out

    def assign_parameters(self) -> None:
        """Assigns the changed parameters to all instances

        This needs to be implemented in each derived object of `BaseTest`

        :raises NotImplementedError: Needs to be implemented in the
            derived object of BaseTest
        """
        for var in self.test_variables:
            for cls in self.list_cls:
                if hasattr(cls, var.name):
                    setattr(cls, var.name, var.value)

    def change_and_assign_single_variable(
        self, var: BaseChangingVariable,
    ) -> None:
        """`var` is changed and its value is assigned so the attribute with
        the name `var.name`.

        The method `var.change` is called which randomly changes
        `var.value`. `var.value` is then assigned to the attribute of `self`
        with the name `var.name`.
        Assigning is handled by :meth:`BaseTest.assign_parameters()`

        :param var: Attribute name
        """
        var.change()
        self.assign_parameters()

    def assertAlmostEqual(
        self,
        val1: Number,
        val2: Number,
        threshold: Optional[float] = 1e-12,
        print_results: Optional[bool] = False,
    ) -> None:

        super().assertAlmostEqual(
            val1, val2, threshold, print_results,
            user_msg=f'\nseed number {self._seed}\n',
        )


if __name__ == '__main__':
    pass



import inspect
import unittest

import numpy as np
from matplotlib import pyplot as plt

from osaft import ARFPlot, Gorkov1962, King1934, WaveTypes, Yosioka1955
from osaft.solutions.base_arf import BaseARF
from osaft.tests.basetest_plotting import BaseTestPlotting


class TestARFPlot(BaseTestPlotting):

    def setUp(self) -> None:
        super().setUp()

        self.plotter = ARFPlot()

    def create_solutions(self) -> list[BaseARF]:
        f = 1e6
        R_0 = 1e-6
        rho_s = 1850
        c_s = 5000
        rho_f = 997
        c_f = 1500
        p_0 = 1e5
        wave_types = WaveTypes.TRAVELLING
        position = None

        out = []
        out.append(
            King1934.ARF(
                f, R_0,
                rho_s,
                rho_f, c_f,
                p_0, wave_types, position,
            ),
        )

        out.append(
            Yosioka1955.ARF(
                f, R_0,
                rho_s, c_s,
                rho_f, c_f,
                p_0, wave_types, position,
            ),
        )

        out.append(
            Gorkov1962.ARF(
                f, R_0,
                rho_s, c_s,
                rho_f, c_f,
                p_0, wave_types, position,
            ),
        )

        return out

    def test_warning(self) -> list[BaseARF]:
        solutions = self.create_solutions()
        for sol in solutions:
            self.plotter.add_solution(sol)
            self.assertRaises(ValueError, self.plotter.add_solution, sol)

    def test_modulo_in_linestyles(self) -> None:
        solutions = self.create_solutions()
        for i in np.arange(6):
            for sol in solutions:
                tmp = sol
                tmp.name = f'{tmp.name}_{i}'
                self.plotter.add_solution(tmp)

        self.plotter.set_abscissa(np.linspace(1e-6, 1e-5), 'R_0')

        fig, ax = self.plotter.plot_solutions()

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_add_and_remove_solutions(self) -> None:
        solutions = self.create_solutions()
        for sol in solutions:
            self.plotter.add_solution(sol)
        self.plotter.remove_solution(solutions[1])

    def test_repeated_ARF_over_R(self) -> None:
        solutions = self.create_solutions()

        self.plotter.set_abscissa(np.linspace(1e-6, 1e-5), 'R_0')

        for sol in solutions:
            self.plotter.add_solution(sol)

        fig, ax = self.plotter.plot_solutions()
        fig, ax = self.plotter.plot_solutions()

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_ARF_over_R(self) -> None:
        solutions = self.create_solutions()

        self.plotter.set_abscissa(np.linspace(1e-6, 1e-5, num=100), 'R_0')

        for sol in solutions:
            self.plotter.add_solution(sol)

        fig, ax = self.plotter.plot_solutions()

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_normARF_over_f(self) -> None:
        solutions = self.create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), 'f')

        for sol in solutions:
            self.plotter.add_solution(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization_name=solutions[1].name,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_normARF_over_f_semilogx(self) -> None:
        solutions = self.create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), 'f')

        for sol in solutions:
            self.plotter.add_solution(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization_name=solutions[1].name,
            plot_method=plt.semilogx,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_normARF_over_f_loglog(self) -> None:
        solutions = self.create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), 'f')

        for sol in solutions:
            self.plotter.add_solution(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization_name=solutions[1].name,
            plot_method=plt.loglog,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Union

from osaft import log
from osaft.core.backgroundfields import WaveTypes
from osaft.core.frequency import Frequency
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import exp, full_range
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_scattering import BaseScatteringRigidParticle
from osaft.solutions.Doinikov1994Rigid.base import BaseDoinikov1994Rigid


class ScatteringField(BaseDoinikov1994Rigid, BaseScatteringRigidParticle):
    """Scattering field class for Doinikov (viscous fluid-rigid sphere; 1994)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param position: Position in the standing wave field [rad]
    :param N_max: Highest order mode
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float,
        rho_f: float, c_f: float,
        eta_f: float, zeta_f: float,
        p_0: float,
        wave_type: Optional[WaveTypes] = WaveTypes.STANDING,
        position: Optional[float] = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method
        """

        # init of parent class
        BaseDoinikov1994Rigid.__init__(
            self, frequency, radius, rho_s, rho_f,
            c_f, eta_f, zeta_f, p_0,
            wave_type, position, N_max,
        )
        BaseScatteringRigidParticle.__init__(self, N_max)

        if type(self) is ScatteringField:
            log.debug(repr(self))
            log.info(str(self))

        # Dependent variables
        self._alpha_n = ActiveVariable(
            self._reset_list,
            'scattering coefficient alpha_n',
        )

        self._beta_n = ActiveVariable(
            self._reset_list,
            'scattering coefficient beta_n',
        )

        self._gamma_n = ActiveVariable(self._reset_list)
        self._xi_n = ActiveVariable(self._reset_list)

        self._mu_1 = ActiveVariable(self._compute_mu_1)
        self._mu_2 = ActiveVariable(self._compute_mu_2)
        self._mu_3 = ActiveVariable(self._compute_mu_3)
        self._mu_4 = ActiveVariable(self._compute_mu_4)

        self._x = ActiveVariable(self._compute_x)
        self._x_v = ActiveVariable(self._compute_x_v)

        # define dependencies
        self._x.is_computed_by(
            self.fluid._k_f, self.sphere._R_0,
        )
        self._x_v.is_computed_by(
            self.fluid._k_v, self.sphere._R_0,
        )

        self._mu_1.is_computed_by(
            self._rho_t, self._x,
        )
        self._mu_2.is_computed_by(
            self._rho_t, self._x,
        )
        self._mu_3.is_computed_by(
            self._rho_t, self._x_v,
        )
        self._mu_4.is_computed_by(
            self._rho_t, self._x_v, self._x,
            self._mu_2, self._mu_3,
        )

        self._gamma_n.is_computed_by(self._x_v)

        self._xi_n.is_computed_by(
            self._x, self._x_v, self._gamma_n,
        )

        self._alpha_n.is_computed_by(
            self._gamma_n, self._xi_n, self._x, self._x_v,
            self._mu_1, self._mu_3, self._mu_4,
        )

        self._beta_n.is_computed_by(
            self._gamma_n, self._xi_n, self._x, self._x_v,
            self._mu_1, self._mu_2, self._mu_4,
        )

    def __repr__(self):
        return (
            f'Donikov1994Rigid.ScatteringFiels(frequency={self.f}, '
            f'radius={self.R_0}, '
            f'rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, '
            f'eta_f={self.eta_f}, zeta_f={self.zeta_f}, '
            f'p_0={self.p_0}, position={self.position}, {self.wave_type}, '
            f'N_max={self.N_max})'
        )

    def __str__(self):
        out = 'Doinikov\'s(1994 rigid - viscous) solution '
        out += 'to the scattering field'
        out += ' with following properties: \n'
        out += SF.get_str_text('Wave type', '', self.wave_type, '')
        out += SF.get_str_text('Frequency', 'f', self.f, 'Hz')
        out += SF.get_str_text('Pressure', 'p_0', self.p_0, 'Pa')
        out += SF.get_str_text(
            'Position', 'd',
            self.position, 'rad',
        )
        out += SF.get_str_text(
            'Wavelength', 'lambda',
            self.field.lambda_f, 'm',
        )

        out += 'Fluid\n'
        out += SF.get_str_text(
            'Density', 'rho_f',
            self.rho_f, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_f, 'm/s',
        )
        out += SF.get_str_text(
            'Viscosity', 'eta_f',
            self.eta_f, 'Pa s',
        )
        out += SF.get_str_text(
            'Bulk Viscosity', 'zeta_f',
            self.zeta_f, 'Pa s',
        )

        out += 'Particle\n'
        out += SF.get_str_text(
            'Radius', 'R_0',
            self.R_0, 'm',
        )
        out += SF.get_str_text(
            'Density', 'rho_s',
            self.rho_s, 'kg/m^3',
        )
        out += 'Other\n'
        out += SF.get_str_text(
            '#modes', 'N_max',
            self.N_max, None,
        )
        return out
    # -----------------------------------------------------
    # Dependent Variables
    # -----------------------------------------------------

    def xi_n(self, n: int) -> complex:
        """ coefficient :math:`\\xi_n` (3.22)

        :param n: order
        """
        if n < len(self._xi_n.value):
            return self._xi_n.value[n]
        else:
            self._compute_xi_n(n)
            return self._xi_n.value[n]

    def _compute_xi_n(self, N: int) -> None:
        # computation according to (3.13) and (3.20)

        N_old = len(self._xi_n.value)
        for n in full_range(N_old, N):
            new_value = -n * (n + 1)
            new_value *= Sp.hankelh1(n, self.x) * Sp.hankelh1(n, self.x_v)
            new_value += self.x * Sp.d1_hankelh1(n, self.x) * self.gamma_n(n)

            self._xi_n.value.append(new_value)

    def gamma_n(self, n: int) -> complex:
        """ coefficient :math:`\\gamma_n` (3.22)

        :param n: order
        """
        if n < len(self._gamma_n.value):
            return self._gamma_n.value[n]
        else:
            self._compute_gamma_n(n)
            return self._gamma_n.value[n]

    def _compute_gamma_n(self, N: int) -> None:
        # computation according to (3.13) and (3.20)

        N_old = len(self._gamma_n.value)
        for n in full_range(N_old, N):
            new_value = Sp.hankelh1(n, self.x_v)
            new_value += self.x_v * Sp.d1_hankelh1(n, self.x_v)

            self._gamma_n.value.append(new_value)

    def alpha_n(self, n: int) -> complex:
        """ coefficient :math:`\\alpha_n` (3.13) and (3.20)

        :param n: order
        """
        if n < len(self._alpha_n.value):
            return self._alpha_n.value[n]
        else:
            self._compute_alpha_n(n)
            return self._alpha_n.value[n]

    def _compute_alpha_n(self, N: int) -> None:
        # computation according to (3.13), (3.14) and (3.20)

        N_old = len(self._alpha_n.value)
        for n in full_range(N_old, N):
            if n == 0:
                new_value = -Sp.besselj(1, self.x) / Sp.hankelh1(1, self.x)
            elif n == 1:
                new_value = -2 * (1 - self.rho_t**2)
                new_value *= Sp.besselj(1, self.x) * Sp.hankelh1(1, self.x_v)
                new_value -= self.mu_1 * self.mu_3
                new_value /= self.mu_4
            else:
                new_value = n * (n + 1) * Sp.besselj(n, self.x)
                new_value *= Sp.hankelh1(n, self.x_v)

                new_value -= self.x * (
                    self.gamma_n(n) *
                    Sp.d1_besselj(n, self.x)
                )
                new_value /= self.xi_n(n)

            self._alpha_n.value.append(new_value)

    def beta_n(self, n: int) -> complex:
        """ coefficient :math:`\\beta_n` (3.13) and (3.20)

        :param n: order
        """
        if n < len(self._beta_n.value):
            return self._beta_n.value[n]
        else:
            self._compute_beta_n(n)
            return self._beta_n.value[n]

    def _compute_beta_n(self, N: int) -> None:
        # computation according to (3.13), (3.15) and (3.21)

        N_old = len(self._beta_n.value)
        for n in full_range(N_old, N):
            if n == 0:
                new_value = 0.0
            elif n == 1:
                new_value = self.mu_1 * Sp.hankelh1(1, self.x)
                new_value -= self.mu_2 * Sp.besselj(1, self.x)

                new_value *= -(1 - self.rho_t)
                new_value /= self.mu_4
            else:
                new_value = Sp.d1_besselj(n, self.x) * Sp.hankelh1(n, self.x)
                new_value -= Sp.besselj(n, self.x) * Sp.d1_hankelh1(n, self.x)
                new_value *= self.x
                new_value /= -self.xi_n(n)

            self._beta_n.value.append(new_value)

    @staticmethod
    def _reset_list() -> list:
        return []

    @property
    def x(self) -> complex:
        """Product of :attr:`~.k_f` and :attr:`~.R_0`
        """
        return self._x.value

    def _compute_x(self) -> complex:
        return self.k_f * self.R_0

    @property
    def x_v(self) -> complex:
        """Product of :attr:`~.k_v` and :attr:`~.R_0`
        """
        return self._x_v.value

    def _compute_x_v(self) -> complex:
        return self.k_v * self.R_0

    @property
    def mu_1(self) -> complex:
        """ :math:`\\mu_1` according to (3.16)
        """
        return self._mu_1.value

    def _compute_mu_1(self) -> complex:
        out = self.rho_t * Sp.besselj(1, self.x)
        out -= self.x * Sp.d1_besselj(1, self.x)
        return out

    @property
    def mu_2(self) -> complex:
        """ :math:`\\mu_2` according to (3.17)
        """
        return self._mu_2.value

    def _compute_mu_2(self) -> complex:
        out = self.rho_t * Sp.hankelh1(1, self.x)
        out -= self.x * Sp.d1_hankelh1(1, self.x)
        return out

    @property
    def mu_3(self) -> complex:
        """ :math:`\\mu_3` according to (3.18)
        """
        return self._mu_3.value

    def _compute_mu_3(self) -> complex:
        out = (1 + 2 * self.rho_t) * Sp.hankelh1(1, self.x_v)
        out += self.x_v * Sp.d1_hankelh1(1, self.x_v)
        return out

    @property
    def mu_4(self) -> complex:
        """ :math:`\\mu_4` according to (3.19)
        """
        return self._mu_4.value

    def _compute_mu_4(self) -> complex:
        out = 2 * (1 - self.rho_t**2)
        out *= Sp.hankelh1(1, self.x)
        out *= Sp.hankelh1(1, self.x_v)
        out += self.mu_2 * self.mu_3
        return out

    # -----------------------------------------------------
    # Methods
    # -----------------------------------------------------

    def V_r_sc(self, n: int, r: float) -> complex:
        """Radial scattering field velocity term of mode `n`
        without Legendre coefficients

        Returns radial scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        out = self.k_f * self.alpha_n(n) * Sp.d1_hankelh1(n, r * self.k_f)
        out -= n * (n + 1) / r * self.beta_n(n) * Sp.hankelh1(n, r * self.k_v)
        out *= self.A_in(n)
        return out

    def V_theta_sc(self, n: int, r: float) -> complex:
        """ Tangential scattering field velocity term of mode n
        without Legendre coefficients

        Returns tangential scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """

        arg = self.k_v * r

        out = Sp.hankelh1(n, arg)
        out += arg * Sp.d1_hankelh1(n, arg)
        out *= -self.beta_n(n)

        out += self.alpha_n(n) * Sp.hankelh1(n, self.k_f * r)

        out *= self.A_in(n) / r

        return out

    def particle_velocity(self, t: float) -> complex:
        """Particle velocity

        Returns the value of the particle velocity
        in the direction of the axis of rotational
        symmetry of the radiation field in [m/s]

        :param t: time [s]
        """

        out = Sp.besselj(1, self.x) + self.alpha_n(1) * Sp.hankelh1(1, self.x)
        out += 2 * self.beta_n(1) * Sp.hankelh1(1, self.x_v)
        out *= self.rho_t * self.A_in(1) * self.k_f
        out /= self.x
        out *= exp(-1j * self.omega * t)
        return out


if __name__ == '__main__':
    pass



from __future__ import annotations

from typing import Optional, Union

from osaft.core.backgroundfields import BackgroundField, WaveTypes
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.fluids import InviscidFluid
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_solution import BaseSolution


class BaseYosioka(BaseSphereFrequencyComposite, BaseSolution):
    """Base class for Yosioka & Kawasima (1955)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the particle [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param c_s: Speed of sound of the particle [m/s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound in the fluid [m/s]
    :param p_0: Pressure amplitude of the incident field [Pa]
    :param position: Position in the standing wave field [rad]
    :param wave_type: Either in or progressive wave [-]
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float, c_s: float,
        rho_f: float, c_f: float,
        p_0: float, wave_type: WaveTypes,
        position: Optional[float] = None,
    ) -> None:

        """Constructor method
        """

        # setting default name
        self._name = 'Yosioka1955'

        # init of parent class
        super().__init__(frequency, radius)

        # Initialize Components
        self.scatterer = InviscidFluid(self.frequency, rho_s, c_s)
        self.fluid = InviscidFluid(self.frequency, rho_f, c_f)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        # Dependent variables
        self._labmda_rho = ActiveVariable(
            self._compute_lambda_rho,
            'ratio of densities',
        )
        self._sigma = ActiveVariable(
            self._compute_sigma,
            'ratio of speed of sounds',
        )

        self._x_f = ActiveVariable(
            self._compute_x_f,
            'dimensionless fluid wavenumber',
        )
        self._x_s = ActiveVariable(
            self._compute_x_s,
            'dimensionless particle wavenumber',
        )

        # Dependencies
        self._labmda_rho.is_computed_by(
            self.scatterer._rho_f,
            self.fluid._rho_f,
        )
        self._sigma.is_computed_by(self.scatterer._c_f, self.fluid._c_f)
        self._x_f.is_computed_by(self.sphere._R_0, self.fluid._k_f)
        self._x_s.is_computed_by(self.sphere._R_0, self.scatterer._k_f)

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def sigma(self) -> float:
        """ratio of speeds of sound [-].
         """
        return self._sigma.value

    @property
    def lambda_rho(self) -> float:
        """ratio of densities [-]
         """
        return self._labmda_rho.value

    @property
    def x_f(self) -> float:
        """ dimensionless wavenumber in the fluid [-]
        """
        return self._x_f.value

    @property
    def x_s(self) -> float:
        """ dimensionless wavenumber in the particle
        """
        return self._x_s.value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to :attr:`osaft.core.backgroundfields.BackgroundField.p_0`
        """
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> str:
        """ Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: str) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`
        """
        return self.scatterer.rho_f

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_f = value

    @property
    def c_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`
        """
        return self.scatterer.c_f

    @c_s.setter
    def c_s(self, value: float) -> None:
        self.scatterer.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`
        """
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`
        """
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.kappa_f`
        """
        return self.fluid.kappa_f

    @property
    def k_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.k_f`
        """
        return self.fluid.k_f

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def k_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.k_f`
        """
        return self.scatterer.k_f

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def I_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.acoustic_intensity`
        """
        return self.field.I_ac

    @property
    def E_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.E_ac`
        """
        return self.field.E_ac

    def A_in(self, n) -> complex:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.acoustic_intensity`
        """
        return self.field.A_in(n)

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    @staticmethod
    def _reset_coeff() -> list:
        return []

    def _compute_sigma(self) -> float:
        """Compute sigma (ratio of speed of sound)
        """
        return self.c_s / self.c_f

    def _compute_lambda_rho(self) -> float:
        """Compute lambda (ratio of densities)
        """
        return self.rho_s / self.rho_f

    def _compute_x_f(self) -> float:
        """Compute x_f
        """
        return self.R_0 * self.k_f

    def _compute_x_s(self) -> float:
        """Compute x_s
        """
        return self.R_0 * self.k_s


if __name__ == '__main__':
    pass



import unittest

import numpy as np

from osaft.core.functions import SpecialFunctions as Sp
from osaft.solutions.Doinikov1994Compressible.scattering import ScatteringField
from osaft.tests.basetest import BaseTest
from osaft.tests.basetest_scattering import HelperScattering


class TestScattering(HelperScattering, BaseTest):

    def setUp(self) -> None:
        super().setUp()

        self._v_particle_threshold = 1e-5
        self._v_fluid_threshold = 1e-3
        self._v_boundary_conditions = 1e-10

        self.cls = ScatteringField(
            self.f, self.R_0,
            self.rho_s, self.c_s,
            self.eta_p, self.zeta_p,
            self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
            self.p_0, self.wave_type,
            self.position,
        )

    # ------------------
    # fluid velocity
    # ------------------

    def V_r_sc(self, n, r) -> complex:
        # grad(phi)
        out = self.cls.k_f * self.alpha_n(n)
        out *= Sp.d1_hankelh1(n, self.cls.k_f * r)
        # curl(Psi)
        arg = self.cls.k_v * r
        out -= n * (n + 1) / r * self.beta_n(n) * Sp.hankelh1(n, arg)

        return out

    def V_theta_sc(self, n, r) -> complex:
        # curl(Psi)
        arg_v = self.cls.k_v * r
        out = Sp.hankelh1(n, arg_v)
        out += arg_v * Sp.d1_hankelh1(n, arg_v)
        out *= -self.beta_n(n)
        # grad(phi)
        out += self.alpha_n(n) * Sp.hankelh1(n, self.cls.k_f * r)

        out /= r
        return out

    # ------------------
    # particle velocity
    # ------------------

    def radial_particle_velocity(self, r, theta, t, mode):
        def vel(n, r):
            out = self.cls.k_s * self.alpha_hat_n(n)
            out *= Sp.d1_besselj(n, r * self.cls.k_s)

            arg = r * self.cls.k_vs
            out -= n * (n + 1) / r * self.beta_hat_n(n) * Sp.besselj(n, arg)
            return out

        out = self.cls.radial_mode_superposition(
            vel, r, theta, t, mode,
        )

        return out

    def tangential_particle_velocity(self, r, theta, t, mode):
        def vel(n, r):
            arg = r * self.cls.k_vs
            out = Sp.besselj(n, arg) + arg * Sp.d1_besselj(n, arg)
            out *= -self.beta_hat_n(n)

            out += self.alpha_hat_n(n) * Sp.besselj(n, r * self.cls.k_s)

            return out / r

        out = self.cls.tangential_mode_superposition(
            vel, r, theta, t, mode,
        )

        return out

    # ------------------
    # generic test method for coefficients
    # ------------------

    def test_coefficients(self) -> None:
        dict_of_methods = {}
        names = [
            'n_1', 'n_2', 'n_3', 'n_4',
            'm_11', 'm_12', 'm_13', 'm_14',
            'm_21', 'm_22', 'm_23', 'm_24',
            'm_31', 'm_32', 'm_33', 'm_34',
            'm_41', 'm_42', 'm_43', 'm_44',
            'det_M_n', 'det_M_1_n', 'det_M_2_n', 'det_M_3_n', 'det_M_4_n',
            'alpha_n', 'beta_n', 'alpha_hat_n', 'beta_hat_n',
        ]
        for name in names:
            dict_of_methods[name] = None

        self._test_methods_n(dict_of_methods, threshold=1e-8)

    # ------------------
    # testing scattering coefficients
    # ------------------

    def alpha_n(self, n: int) -> complex:
        return self.det_M_1_n(n) / self.det_M_n(n)

    def beta_n(self, n: int) -> complex:
        if n > 0:
            return self.det_M_2_n(n) / self.det_M_n(n)
        else:
            return 0

    def alpha_hat_n(self, n: int) -> complex:
        return self.det_M_3_n(n) / self.det_M_n(n)

    def beta_hat_n(self, n: int) -> complex:
        if n > 0:
            return self.det_M_4_n(n) / self.det_M_n(n)
        else:
            return 0

    # ------------------
    # testing determinants
    # ------------------

    def _get_matrix_from_list(self, coefficients: list, n: int) -> np.ndarray:
        out = np.zeros((4, 4), dtype=np.complex_)
        for i, name in enumerate(coefficients):
            col = i % 4
            row = i // 4
            #  print(name, i, row, col)
            out[row, col] = getattr(self, name)(n)

        return out

    def _get_M_matrix(self, n: int) -> np.ndarray:
        names = [
            'm_11', 'm_12', 'm_13', 'm_14',
            'm_21', 'm_22', 'm_23', 'm_24',
            'm_31', 'm_32', 'm_33', 'm_34',
            'm_41', 'm_42', 'm_43', 'm_44',
        ]

        return self._get_matrix_from_list(names, n)

    def _get_M_i_matrix(self, n: int, i: int) -> np.ndarray:
        M = self._get_M_matrix(n)
        for j in np.arange(4):
            M[j, i - 1] = getattr(self, f'n_{j+1:1.0f}')(n)
        return M

    def det_M_n(self, n: int) -> complex:
        return np.linalg.det(self._get_M_matrix(n))

    def det_M_1_n(self, n: int) -> complex:
        return np.linalg.det(self._get_M_i_matrix(n, 1))

    def det_M_2_n(self, n: int) -> complex:
        return np.linalg.det(self._get_M_i_matrix(n, 2))

    def det_M_3_n(self, n: int) -> complex:
        return np.linalg.det(self._get_M_i_matrix(n, 3))

    def det_M_4_n(self, n: int) -> complex:
        return np.linalg.det(self._get_M_i_matrix(n, 4))

    # ------------------
    # testing m_1i
    # ------------------

    def m_11(self, n: int) -> complex:
        out = self.x * Sp.d1_hankelh1(n, self.x)
        return out

    def m_12(self, n: int) -> complex:
        out = -n * (n + 1) * Sp.hankelh1(n, self.x_v)
        return out

    def m_13(self, n: int) -> complex:
        out = -self.x_hat * Sp.d1_besselj(n, self.x_hat)
        return out

    def m_14(self, n: int) -> complex:
        out = n * (n + 1) * Sp.besselj(n, self.x_hat_v)
        return out

    # ------------------
    # testing m_2i
    # ------------------

    def m_21(self, n: int) -> complex:
        out = Sp.hankelh1(n, self.x)
        return out

    def m_22(self, n: int) -> complex:
        out = -Sp.hankelh1(n, self.x_v)
        out -= self.x_v * Sp.d1_hankelh1(n, self.x_v)
        return out

    def m_23(self, n: int) -> complex:
        out = -Sp.besselj(n, self.x_hat)
        return out

    def m_24(self, n: int) -> complex:
        out = Sp.besselj(n, self.x_hat_v)
        out += self.x_hat_v * Sp.d1_besselj(n, self.x_hat_v)
        return out

    # ------------------
    # testing m_3i
    # ------------------

    def m_31(self, n: int) -> complex:
        out = 1j * self.cls.rho_f * self.cls.c_f**2 / self.cls.omega
        out += self.cls.zeta_f
        out -= 2 * self.cls.eta_f / 3
        out *= -Sp.hankelh1(n, self.x)

        out += 2 * self.cls.eta_f * Sp.d2_hankelh1(n, self.x)

        out *= self.x**2

        return out

    def m_32(self, n: int) -> complex:
        out = Sp.hankelh1(n, self.x_v)
        out -= self.x_v * Sp.d1_hankelh1(n, self.x_v)

        out *= 2 * n * (n + 1) * self.cls.eta_f
        return out

    def m_33(self, n: int) -> complex:
        out = 1j * self.cls.rho_s * self.cls.c_s**2 / self.cls.omega
        out += self.cls.zeta_s
        out -= 2 * self.cls.eta_s / 3
        out *= Sp.besselj(n, self.x_hat)

        out -= 2 * self.cls.eta_s * Sp.d2_besselj(n, self.x_hat)

        out *= self.x_hat**2
        return out

    def m_34(self, n: int) -> complex:
        out = -Sp.besselj(n, self.x_hat_v)
        out += self.x_hat_v * Sp.d1_besselj(n, self.x_hat_v)

        out *= 2 * n * (n + 1) * self.cls.eta_s
        return out

    # ------------------
    # testing m_4i
    # ------------------

    def m_41(self, n: int) -> complex:
        out = self.x * Sp.d1_hankelh1(n, self.x)
        out -= Sp.hankelh1(n, self.x)

        out *= 2 * self.cls.eta_f

        return out

    def m_42(self, n: int) -> complex:
        out = (n ** 2 + n - 2) * Sp.hankelh1(n, self.x_v)

        out += self.x_v**2 * Sp.d2_hankelh1(n, self.x_v)

        out *= -self.cls.eta_f
        return out

    def m_43(self, n: int) -> complex:
        out = Sp.besselj(n, self.x_hat)
        out -= self.x_hat * Sp.d1_besselj(n, self.x_hat)

        out *= 2 * self.cls.eta_s
        return out

    def m_44(self, n: int) -> complex:
        out = (n ** 2 + n - 2) * Sp.besselj(n, self.x_hat_v)

        out += self.x_hat_v**2 * Sp.d2_besselj(n, self.x_hat_v)

        out *= self.cls.eta_s
        return out

    # ------------------
    # testing n_i
    # ------------------

    def n_1(self, n: int) -> complex:
        out = -self.x * Sp.d1_besselj(n, self.x)
        out *= self.cls.A_in(n)
        return out

    def n_2(self, n: int) -> complex:
        out = -Sp.besselj(n, self.x)
        out *= self.cls.A_in(n)
        return out

    def n_3(self, n: int) -> complex:
        out = 1j * self.cls.rho_f * self.cls.c_f**2 / self.cls.omega
        out += self.cls.zeta_f
        out -= 2 * self.cls.eta_f / 3
        out *= Sp.besselj(n, self.x)

        out -= 2 * self.cls.eta_f * Sp.d2_besselj(n, self.x)

        out *= self.x ** 2 * self.cls.A_in(n)

        return out

    def n_4(self, n: int) -> complex:
        out = Sp.besselj(n, self.x)
        out -= self.x * Sp.d1_besselj(n, self.x)

        out *= 2 * self.cls.eta_f * self.cls.A_in(n)
        return out

    # ------------------
    @property
    def x(self):
        return self.cls.x

    @property
    def x_v(self):
        return self.cls.x_v

    @property
    def x_hat(self):
        return self._compute_x_hat()

    def _compute_x_hat(self) -> complex:
        return self.R_0 * self.cls.k_s

    def test_x(self) -> None:
        self.do_testing(
            func_1=self._compute_x_hat,
            func_2=lambda: self.cls.x_hat,
        )

    # ------------------

    @property
    def x_hat_v(self):
        return self._compute_x_hat_v()

    def _compute_x_hat_v(self) -> complex:
        return self.R_0 * self.cls.k_vs

    def test_x_hat_v(self) -> None:
        self.do_testing(
            func_1=self._compute_x_hat_v,
            func_2=lambda: self.cls.x_hat_v,
        )


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Union

from osaft.core.backgroundfields import WaveTypes
from osaft.core.fluids import ViscousFluid
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.solutions.Doinikov1994Rigid.base import BaseDoinikov1994Rigid


class BaseDoinikov1994Compressible(BaseDoinikov1994Rigid):
    """Base class for Doinikov (viscous fluid-viscous sphere; 1994)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the solid [m]
    :param rho_s: Density of the fluid-like solid [kg/m^3]
    :param c_s: Speed of sound of the sphere [m/s]
    :param eta_s: shear viscosity of the sphere [Pa s]
    :param zeta_s: bulk viscosity of the sphere [Pa s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param position: Position within the standing wave field [m]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param N_max: Highest order mode
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float, c_s: float, eta_s: float, zeta_s: float,
        rho_f: float, c_f: float, eta_f: float, zeta_f: float,
        p_0: float, wave_type: WaveTypes, position: float,
        N_max: Optional[int] = 3,
    ) -> None:
        """Constructor method
        """

        super().__init__(
            frequency, radius, rho_s, rho_f, c_f, eta_f, zeta_f,
            p_0, position, wave_type, N_max,
        )

        self.scatterer = ViscousFluid(
            self.frequency, rho_s, c_s, eta_s,
            zeta_s,
        )

    def __str__(self):
        out = 'Doinikovs\'s  (1994) model (viscous fluid-viscous sphere)'
        out += ' with the following properties: \n'
        out += SF.get_str_text('Frequency', 'f', self.f, 'Hz')
        out += SF.get_str_text('Pressure', 'p_0', self.p_0, 'Pa')
        out += SF.get_str_text('Wavetype', 'wave_type', self.wave_type, None)
        out += SF.get_str_text(
            'Position', 'd',
            self.position, 'Pa',
        )
        out += SF.get_str_text(
            'Wavelength', 'lambda',
            self.field.lambda_f, 'm',
        )

        out += 'Fluid\n'
        out += SF.get_str_text(
            'Density', 'rho_f',
            self.rho_f, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_f, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_f',
            self.kappa_f, '1/Pa',
        )
        out += SF.get_str_text(
            'Shear viscosity', 'eta_f',
            self.eta_f, '1/Pa',
        )
        out += SF.get_str_text(
            'Bulk viscosity', 'zeta_f',
            self.zeta_f, '1/Pa',
        )

        out += 'Particle\n'
        out += SF.get_str_text(
            'Radius', 'R_0',
            self.R_0, 'm',
        )
        out += SF.get_str_text(
            'Density', 'rho_s',
            self.rho_s, 'kg/m^3',
        )
        out += SF.get_str_text(
            'Sound Speed', 'c_0',
            self.c_s, 'm/s',
        )
        out += SF.get_str_text(
            'Compressibility', 'kappa_s',
            self.kappa_s, '1/Pa',
        )
        out += SF.get_str_text(
            'Shear viscosity', 'eta_s',
            self.eta_s, '1/Pa',
        )
        out += SF.get_str_text(
            'Bulk viscosity', 'zeta_s',
            self.zeta_s, '1/Pa',
        )
        return out

    # -------------------------------------------------------------------------
    # Wrappers for Independent sphere Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_s(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.kappa_f`
        """
        return self.scatterer.kappa_f

    @property
    def k_s(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_f`
        """
        return self.scatterer.k_f

    @property
    def k_vs(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_v`
        """
        return self.scatterer.k_v

    @property
    def rho_s(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.rho_f`
        """
        return self.scatterer.rho_f

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_f = value

    @property
    def c_s(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.c_f`
        """
        return self.scatterer.c_f

    @c_s.setter
    def c_s(self, value: float) -> None:
        self.scatterer.c_f = value

    @property
    def eta_s(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.eta_f`
        """
        return self.scatterer.eta_f

    @eta_s.setter
    def eta_s(self, value: float) -> None:
        self.scatterer.eta_f = value

    @property
    def zeta_s(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.zeta_f`
        """
        return self.scatterer.zeta_f

    @zeta_s.setter
    def zeta_s(self, value: float) -> None:
        self.scatterer.zeta_f = value


if __name__ == '__main__':
    pass



from __future__ import annotations

from typing import Optional, Union

from osaft import log
from osaft.core.backgroundfields import BackgroundField, WaveTypes
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.fluids import ViscousFluid
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.solids import RigidSolid
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.base_solution import BaseSolution


class BaseDoinikov1994Rigid(BaseSphereFrequencyComposite, BaseSolution):
    """Base class for Doinikov (viscous fluid-rigid sphere; 1994)

    :param frequency: Frequency [Hz]
    :param radius: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, travel(l)ing or standing
    :param position: Position in the standing wave field [rad]
    :param N_max: Highest order mode
    """

    def __init__(
        self, frequency: Union[Frequency, float, int],
        radius: Union[Sphere, float, int],
        rho_s: float,
        rho_f: float, c_f: float,
        eta_f: float, zeta_f: float,
        p_0: float,
        wave_type: Optional[WaveTypes] = WaveTypes.STANDING,
        position: Optional[float] = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method
        """

        # define name
        self._name = 'Doinikov1994Rigid'

        # init of parent class
        BaseSphereFrequencyComposite.__init__(self, frequency, radius)

        # Initialize Components
        self.scatterer = RigidSolid(self.frequency, rho_s)
        self.fluid = ViscousFluid(self.frequency, rho_f, c_f, eta_f, zeta_f)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        # Independent variables
        self._N_max = PassiveVariable(N_max, 'number of modes N')

        # Dependent Variables
        self._rho_t = ActiveVariable(
            self._compute_rho_t,
            'ratio of densities',
        )
        self._rho_t.is_computed_by(self.fluid._rho_f, self.scatterer._rho_s)

        if type(self) == BaseDoinikov1994Rigid:
            log.info(repr(self))

    def __repr__(self):
        return (
            f'BaseDoinikov1994Rigid(frequency={self.f}, radius={self.R_0}, '
            f'rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, '
            f'eta_f={self.eta_f}, zeta_f={self.zeta_f}, '
            f'p_0={self.p_0}, position={self.position}, {self.wave_type}, '
            f'N_max={self.N_max})'
        )

    @property
    def N_max(self):
        """Cutoff mode number for infinite sums

        :getter: returns number of infinite sum term
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._N_max.value

    @N_max.setter
    def N_max(self, value):
        self._N_max.value = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.p_0`
        """
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> WaveTypes:
        """ Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: WaveTypes) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """ Wraps to :attr:`osaft.core.solids.RigidSolid.rho_s`
        """
        return self.scatterer.rho_s

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_s = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.rho_f`
        """
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.c_f`
        """
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    @property
    def eta_f(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.eta_f`
        """
        return self.fluid.eta_f

    @eta_f.setter
    def eta_f(self, value: float) -> None:
        self.fluid.eta_f = value

    @property
    def zeta_f(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.zeta_f`
        """
        return self.fluid.zeta_f

    @zeta_f.setter
    def zeta_f(self, value: float) -> None:
        self.fluid.zeta_f = value
    # -------------------------------------------------------------------------
    # Wrappers for Dependent Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_t(self) -> float:
        """ Returns the ratio of the densities
        :math:`\\tilde{\\rho}=\\frac{\\rho_f}{\\rho_s}`
        """
        return self._rho_t.value

    def _compute_rho_t(self) -> float:
        return self.rho_f / self.rho_s

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.kappa_f`
        """
        return self.fluid.kappa_f

    @property
    def k_f(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_f`
        """
        return self.fluid.k_f

    @property
    def k_v(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_v`
        """
        return self.fluid.k_v

    @property
    def delta(self) -> float:
        """ Wraps to :attr:`osaft.core.fluids.ViscousFluid.delta`
        """
        return self.fluid.delta

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def abs_pos(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.abs_pos`
        """
        return self.field.abs_pos

    def A_in(self, n) -> complex:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.A_in`
        """
        return self.field.A_in(n)


if __name__ == '__main__':
    pass



from osaft import log
from osaft.core.functions import pi
from osaft.core.variable import ActiveVariable, PassiveVariable


class Sphere:
    """Sphere class

    :param radius: radius in [m]
    """

    def __init__(self, radius: float) -> None:
        """Constructor method
        """

        # Independent variables
        self._R_0 = PassiveVariable(radius, 'radius R_0')

        # Dependent variables
        self._A = ActiveVariable(
            self._compute_surface_area,
            'surface area A',
        )
        self._V = ActiveVariable(
            self._compute_volume,
            'volume V',
        )

        # Dependencies
        self._A.is_computed_by(self._R_0)
        self._V.is_computed_by(self._R_0)

        log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return f'Sphere(R={self.R_0})'

    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def R_0(self) -> float:
        """Radius of the sphere [m]

        :getter: returns the value for the radius
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._R_0.value

    @R_0.setter
    def R_0(self, value: float) -> None:
        self._R_0.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def A(self) -> float:
        """Returns the surface area of the sphere [m^2]
        """
        return self._A.value

    @property
    def V(self) -> float:
        """Returns the volume of the sphere [m^3]
        """
        return self._V.value

    # -------------------------------------------------------------------------
    # Methods for Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_surface_area(self) -> float:
        return 4 * pi * self.R_0 ** 2

    def _compute_volume(self) -> float:
        return 4 / 3 * pi * self.R_0 ** 3


if __name__ == '__main__':
    pass



from typing import Union

from osaft import log
from osaft.core.basecomposite import BaseFrequencyComposite
from osaft.core.frequency import Frequency
from osaft.core.functions import pi, sqrt
from osaft.core.variable import ActiveVariable, PassiveVariable


class InviscidFluid(BaseFrequencyComposite):
    """Class for an inviscid fluid

    :param frequency: frequency in [Hz]
    :param rho: density in [km/m^3]
    :param c: speed of sound [m/s]
    """

    def __init__(
        self, frequency: Union[int, float, Frequency],
        rho: float, c: float,
    ) -> None:
        """Constructor method
        """

        super().__init__(frequency)

        # Independent Variables
        self._rho_f = PassiveVariable(rho, 'density rho')
        self._c_f = PassiveVariable(c, 'speed of sound c')

        # Dependent variables
        self._k_f = ActiveVariable(self._compute_k_f, 'wavenumber k_f')
        self._kappa_f = ActiveVariable(
            self._compute_kappa_f,
            'compressibility kappa_f',
        )
        self._lambda_f = ActiveVariable(
            self._compute_lambda_f,
            'wavelength lambda_f',
        )

        # Dependencies
        self._k_f.is_computed_by(self._c_f, self.frequency._omega)
        self._lambda_f.is_computed_by(self._c_f, self.frequency._f)
        self._kappa_f.is_computed_by(self._c_f, self._rho_f)

        if type(self) is InviscidFluid:
            log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return (
            'InviscidFluid('
            f'frequency={self.frequency}, '
            f'rho={self.rho_f}, '
            f'c={self.c_f}'
            ')'
        )

    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------
    @property
    def c_f(self) -> float:
        r"""Wave speed in the fluid :math:`c_f` [m\s].

        :getter: returns the value for the frequency
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._c_f.value

    @c_f.setter
    def c_f(self, value: float) -> None:
        self._c_f.value = value

    @property
    def rho_f(self) -> float:
        """Density of the fluid :math:`\\rho_f` [kg/m^3].

        :getter: returns the value for the density
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._rho_f.value

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self._rho_f.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        r"""Returns the compressibility of the fluid :math:`\kappa_f` [1/Pa]
        """
        return self._kappa_f.value

    @property
    def k_f(self) -> float:
        """Returns the wave number :math:`k_f` [rad s^-1]
        """
        return self._k_f.value

    @property
    def lambda_f(self) -> float:
        """Returns the wavelength  :math:`\\lambda_f` [m]
        """
        return self._lambda_f.value

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_k_f(self) -> float:
        return self.omega / self.c_f

    def _compute_kappa_f(self) -> float:
        return 1 / self.c_f**2 / self.rho_f

    def _compute_lambda_f(self) -> float:
        return self.c_f / self.f


class ViscousFluid(InviscidFluid):
    """Class for a viscous fluid

    :param frequency: frequency in [Hz]
    :param rho: density in [km/m^3]
    :param c: speed of sound [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
        """

    def __init__(
        self, frequency: Union[int, float, Frequency], rho: float,
        c: float, eta_f: float,
        zeta_f: float,
    ):
        """Constructor method
        """

        super().__init__(frequency, rho, c)

        # Independent Variables
        self._eta_f = PassiveVariable(eta_f, 'shear viscosity eta_f')
        self._zeta_f = PassiveVariable(zeta_f, 'bulk viscosity zeta_f')

        # Dependent Variables
        self._k_v = ActiveVariable(self._compute_k_v, 'viscous wavenumber k_v')
        self._delta = ActiveVariable(
            self._compute_delta,
            'boundary layer thickness delta',
        )
        self._lambda_v = ActiveVariable(
            self._compute_lambda_v,
            'viscous wavelength lambda_v',
        )

        # Dependencies
        self._k_v.is_computed_by(
            self.frequency._omega, self._rho_f,
            self._eta_f,
        )

        self._k_f.is_computed_by(self._rho_f, self._eta_f, self._zeta_f)

        self._lambda_v.is_computed_by(self._k_v)
        self._delta.is_computed_by(self._k_v)

        if type(self) is ViscousFluid:
            log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return(
            'ViscousFluid('
            f'frequency={self.frequency}, '
            f'rho={self.rho_f}, '
            f'c={self.c_f}, '
            f'eta_f={self.eta_f}, '
            f'zeta_f={self.zeta_f}'
            ')'
        )
    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def eta_f(self) -> float:
        """Shear viscosity of the fluid :math:`\\eta_f` [Pa s].

        :getter: returns the shear viscosity
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._eta_f.value

    @eta_f.setter
    def eta_f(self, value: float) -> None:
        self._eta_f.value = value

    @property
    def zeta_f(self) -> float:
        """Bulk viscosity of the fluid :math:`\\zeta_f` [Pa s].

        :getter: returns the bulk viscosity
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._zeta_f.value

    @zeta_f.setter
    def zeta_f(self, value: float) -> None:
        self._zeta_f.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def k_f(self) -> complex:
        """Returns the wave number in the fluid :math:`k_f` [1/m]
        """
        return self._k_f.value

    @property
    def k_v(self) -> complex:
        """Returns the viscous wave number in the fluid :math:`k_v` [1/m]
        """
        return self._k_v.value

    @property
    def delta(self) -> float:
        """Returns the boundary layer thickness :math:`\\delta` [m]
        """
        return self._delta.value

    @property
    def lambda_v(self) -> float:
        """Returns the viscous wave length :math:`\\lambda_v` [m]
        """
        return self._lambda_v.value

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_k_f(self) -> complex:
        return (
            self.omega / self.c_f
            / sqrt(
                1 - 1j * self.omega / (self.rho_f * self.c_f ** 2)
                * (self.zeta_f + 4 * self.eta_f / 3),
            )
        )

    def _compute_k_v(self) -> complex:
        if self.eta_f != 0:
            return (1 + 1j) * sqrt(self.rho_f * self.omega / (2 * self.eta_f))
        else:
            return 0

    def _compute_delta(self) -> float:
        if self.k_v != 0:
            return 1 / self.k_v.imag
        else:
            return 0

    def _compute_lambda_v(self) -> float:
        if self.k_v != 0:
            return 2 * pi / self.k_v.real
        else:
            return 0


class ViscoelasticFluid(ViscousFluid):
    """Class for a viscoelastic fluid

    :param frequency: frequency in [Hz]
    :param rho: density in [km/m^3]
    :param c: speed of sound [m/s]
    :param eta_f: shear viscosity of fluid component [Pa s]
    :param zeta_f: bulk viscosity of fluid component [Pa s]
    :param eta_p: shear viscosity of polymer component [Pa s]
    :param zeta_p: bulk viscosity of polymer component [Pa s]
    :param zeta_p: relaxation time [s]
    :param lambda_M: relaxation time of fluid [s]
        """

    def __init__(
        self, frequency: Union[int, float, Frequency],
        rho: float, c: float,
        eta_f: float, eta_p: float, zeta_f: float, zeta_p: float,
        lambda_M: float,
    ):
        """Constructor method
        """

        super().__init__(frequency, rho, c, eta_f, zeta_f)

        # Independent Variables
        self._eta_p = PassiveVariable(
            eta_p,
            'shear viscosity of polymer component '
            'eta_p',
        )
        self._zeta_p = PassiveVariable(
            zeta_p,
            'shear viscosity of polymer component '
            'zeta_p',
        )
        self._lambda_M = PassiveVariable(
            lambda_M,
            'relaxation time lambda_M',
        )

        # Dependent variables
        self._eta_c = ActiveVariable(self._compute_eta_c)
        self._zeta_c = ActiveVariable(self._compute_zeta_c)

        # Dependencies
        self._k_f.is_computed_by(
            self.frequency._omega,
            self._c_f, self._rho_f,
            self._eta_c, self._zeta_c,
        )

        self._k_v.is_computed_by(self._eta_c, self._zeta_c)

        self._eta_c.is_computed_by(
            self.frequency._omega, self._eta_f,
            self._eta_p, self._lambda_M,
        )

        self._zeta_c.is_computed_by(
            self.frequency._omega,
            self._zeta_f, self._zeta_p,
            self._lambda_M,
        )

        # no if clause needed because last in inheritance
        log.debug(f'Creating {self}')

    def __repr__(self) -> str:
        return(
            'ViscoelasticFluid('
            f'frequency={self.frequency}, '
            f'rho={self.rho_f}, '
            f'c={self.c_f}, '
            f'eta_f={self.eta_f}, '
            f'eta_p={self.eta_p}, '
            f'zeta_f={self.zeta_f}, '
            f'zeta_p={self.zeta_p}, '
            f'lambda_M={self.lambda_M} '
            ')'
        )
    # -------------------------------------------------------------------------
    # Setters and Getters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def eta_p(self) -> float:
        """Polymer component shear viscosity :math:`\\eta_p` [Pa s]

        :getter: Returns the polymer component shear viscosity
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._eta_p.value

    @eta_p.setter
    def eta_p(self, value: float) -> None:
        self._eta_p.value = value

    @property
    def zeta_p(self) -> float:
        """Polymer component bulk viscosity :math:`\\zeta_p` [Pa s]

        :getter: Returns the polymer component bulk viscosity
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._zeta_p.value

    @zeta_p.setter
    def zeta_p(self, value: float) -> None:
        self._zeta_p.value = value

    @property
    def lambda_M(self) -> float:
        """Relaxation time of the fluid :math:`\\lambda_M` [s]

        :getter: Returns the relaxation time of the fluid
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._lambda_M.value

    @lambda_M.setter
    def lambda_M(self, value: float) -> None:
        self._lambda_M.value = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def eta_c(self) -> complex:
        """Complex acoustic shear viscosity of the fluid :math:`\\eta_c` [Pa s]
        """
        return self._eta_c.value

    @property
    def zeta_c(self) -> complex:
        """Complex acoustic bulk viscosity of the fluid :math:`\\zeta_c` [Pa s]
        """
        return self._zeta_c.value

    @property
    def k_f(self) -> complex:
        """Returns the wave number in the fluid :math:`k_f` [1/m]
        """
        return self._k_f.value

    @property
    def k_v(self) -> complex:
        """Returns the viscous wave number in the fluid :math:`k_v` [1/m]
        """
        return self._k_v.value

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_eta_c(self) -> complex:
        return self.eta_f + self.eta_p / (1 - 1j * self.omega * self.lambda_M)

    def _compute_zeta_c(self) -> complex:
        return self.zeta_f + self.zeta_p / (
            1 - 1j * self.omega *
            self.lambda_M
        )

    def _compute_k_f(self) -> complex:
        return (
            self.omega / self.c_f
            / sqrt(
                1 - 1j * self.omega / (self.rho_f * self.c_f ** 2)
                * (self.zeta_c + 4 * self.eta_c / 3),
            )
        )

    def _compute_k_v(self) -> complex:
        if self.eta_c != 0:
            return (1 + 1j) * sqrt(self.rho_f * self.omega / (2 * self.eta_c))
        else:
            return 0


if __name__ == '__main__':
    pass



import unittest

from numpy import real

from osaft import Doinikov1994Compressible
from osaft.core.backgroundfields import WaveTypes, WrongWaveTypeError
from osaft.core.functions import pi, sin, sqrt
from osaft.tests.basetest import BaseTest
from osaft.tests.basetest_arf import HelperStandingARF, HelperTravelingARF


class TestARFBase(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.small_viscous_boundary_layer = True
        self.large_viscous_boundary_layer = False
        self.small_particle_limit = True
        self.N_max = 5

        self.cls = Doinikov1994Compressible.ARF(
            self.f,
            self.R_0,
            self.rho_s, self.c_s, self.eta_s, self.zeta_s,
            self.rho_f, self.c_f, self.eta_f, self.zeta_f,
            self.p_0, self.wave_type,
            self.position,
            self.small_viscous_boundary_layer,
            self.large_viscous_boundary_layer,
            self.small_particle_limit,
            self.N_max,
        )

        self.list_cls = [self.cls]

    def assign_parameters(self) -> None:
        super().assign_parameters()

        self.cls.small_viscous_boundary_layer = (
            self.small_viscous_boundary_layer
        )
        self.cls.large_viscous_boundary_layer = (
            self.large_viscous_boundary_layer
        )
        self.cls.small_particle_limit = self.small_particle_limit
        self.cls.N_max = self.N_max

    @property
    def kappa_t(self) -> float:
        return self.rho_f * self.c_f**2 / self.rho_s / self.c_s**2

    @property
    def mu(self) -> float:
        return sqrt(self.rho_f * self.eta_f / self.rho_s / self.eta_s)

    @property
    def rho_t(self) -> float:
        return self.rho_f / self.rho_s

    @property
    def x_0(self) -> complex:
        return real(self.cls.field.k_f) * self.R_0

    @property
    def norm_delta(self) -> complex:
        return self.cls.delta / self.R_0

    def compute_g1(self):
        return self.compute_g5() * self.compute_g6()

    def compute_g2(self):
        out = 209 + 148 * self.compute_eta_t() + 48 / self.compute_eta_t()
        out *= - 1 / 9 / self.compute_zeta_t()

        out += 19 + 38 * self.compute_eta_t() - 12 / self.compute_eta_t()

        return out

    def compute_g3(self):
        out = 25 + 370 * self.compute_eta_t() - 80 / self.compute_eta_t()
        out *= 1 / 9 / self.compute_zeta_t()

        out += -12 - 19 * self.compute_eta_t() - 4 / self.compute_eta_t()

        return out

    def compute_g4(self):
        out = 5 + 74 * self.compute_eta_t() - 16 / self.compute_eta_t()
        out *= 1 / 9 / self.compute_zeta_t()

        out += -3 - 4 / self.compute_eta_t()

        return out

    def compute_g5(self):
        out = (1 - self.compute_eta_t())
        out *= (19 + 16 / self.compute_eta_t())
        return out

    def compute_g6(self):
        return 89 + 48 / self.compute_eta_t() + 38 * self.compute_eta_t()

    def compute_g7(self):
        return self.compute_g6()**2

    def compute_G(self):
        return (self.compute_S1() + self.compute_S2() + self.compute_S3()) / 3

    def compute_S1(self):
        return 1 - self.kappa_t + 0.9 * (1 - 1 / self.rho_t)

    def compute_S2(self):
        out = 2 * self.compute_f2() - 1
        out *= 2 * self.compute_f1()

        out -= 4 * self.compute_f4()
        return out

    def compute_S3(self):
        out = 3 - 1 / self.compute_eta_t()
        out *= -120 * (self.compute_f3() - self.compute_f4())

        out += 50 * self.compute_f1() + 43 - 10 * self.kappa_t

        out *= self.rho_s - self.rho_f
        out /= 30 * self.rho_f * (3 + 2 / self.compute_eta_t())

        return out

    def compute_f1(self):
        return self.compute_g1() / self.compute_g7()

    def compute_f2(self):
        if self.compute_g1() == 0:
            return 0
        return self.compute_g2() * self.compute_g5() / 2 / self.compute_g1()

    def compute_f3(self):
        return self.compute_g3() * self.compute_g6() / self.compute_g7() / 10

    def compute_f4(self):
        return self.compute_g4() * self.compute_g6() / self.compute_g7() / 2

    def compute_zeta_t(self):
        return self.zeta_s / self.zeta_f

    def compute_eta_t(self):
        return self.eta_s / self.eta_f


class TestARFMethods(TestARFBase):

    # ------------------

    def test_NotImplemented_solution(self) -> None:
        self.cls.wave_type = 'random'

        self.assertRaises(
            WrongWaveTypeError,
            self.cls.acoustic_radiation_force,
        )
    # ------------------

    def test_small_particle_small_delta_standing_not_implemented(self) -> None:

        self.cls.small_viscous_boundary_layer = True
        self.cls.small_particle_limit = True
        self.cls.wave_type = WaveTypes.STANDING
        self.assertRaises(
            NotImplementedError,
            self.cls.acoustic_radiation_force,
        )

    def test_small_particle_limit(self) -> None:
        self.assign_parameters()
        self.assertEqual(
            self.small_particle_limit,
            self.cls.small_particle_limit,
        )

    def test_small_particle_limit_error(self) -> None:
        self.small_particle_limit = False
        self.assertRaises(NotImplementedError, self.assign_parameters)

    def test_small_viscous_boundary_layer(self) -> None:
        self.assign_parameters()
        self.assertEqual(
            self.small_viscous_boundary_layer,
            self.cls.small_viscous_boundary_layer,
        )
        self.small_viscous_boundary_layer = True
        self.assign_parameters()
        self.assertEqual(
            self.small_viscous_boundary_layer,
            self.cls.small_viscous_boundary_layer,
        )

    def test_small_large_BL_ValueError(self):
        self.wave_type = WaveTypes.STANDING
        self.small_viscous_boundary_layer = True
        self.large_viscous_boundary_layer = True
        self.assign_parameters()

        self.assertRaises(
            ValueError,
            self.cls.acoustic_radiation_force,
        )

    def test_arbitrary_BL_ValueError(self):
        self.wave_type = WaveTypes.STANDING
        self.small_viscous_boundary_layer = False
        self.large_viscous_boundary_layer = False
        self.assign_parameters()
        self.assertRaises(
            NotImplementedError,
            self.cls.acoustic_radiation_force,
        )

    def test_g1(self):
        self.do_testing(
            func_2=self.compute_g1,
            func_1=lambda: self.cls._g1.value,
        )

    def test_g2(self):
        self.do_testing(
            func_2=self.compute_g2,
            func_1=lambda: self.cls._g2.value,
        )

    def test_g3(self):
        self.do_testing(
            func_2=self.compute_g3,
            func_1=lambda: self.cls._g3.value,
        )

    def test_g4(self):
        self.do_testing(
            func_2=self.compute_g4,
            func_1=lambda: self.cls._g4.value,
        )

    def test_g5(self):
        self.do_testing(
            func_2=self.compute_g5,
            func_1=lambda: self.cls._g5.value,
        )

    def test_g6(self):
        self.do_testing(
            func_2=self.compute_g6,
            func_1=lambda: self.cls._g6.value,
        )

    def test_g7(self):
        self.do_testing(
            func_2=self.compute_g7,
            func_1=lambda: self.cls._g7.value,
        )

    def test_G(self):
        self.do_testing(
            func_2=self.compute_G,
            func_1=lambda: self.cls._G.value,
        )

    def test_S1(self):
        self.do_testing(
            func_2=self.compute_S1,
            func_1=lambda: self.cls._S1.value,
        )

    def test_S2(self):
        self.do_testing(
            func_2=self.compute_S2,
            func_1=lambda: self.cls._S2.value,
        )

    def test_S3(self):
        self.do_testing(
            func_2=self.compute_S3,
            func_1=lambda: self.cls._S3.value,
        )

    def test_f1(self):
        self.do_testing(
            func_2=self.compute_f1,
            func_1=lambda: self.cls._f1.value,
        )

    def test_f2(self):
        self.do_testing(
            func_2=self.compute_f2,
            func_1=lambda: self.cls._f2.value,
        )

    def test_f3(self):
        self.do_testing(
            func_2=self.compute_f3,
            func_1=lambda: self.cls._f3.value,
        )

    def test_f4(self):
        self.do_testing(
            func_1=self.compute_f4,
            func_2=lambda: self.cls._f4.value,
        )

    def test_zeta_t(self):
        self.do_testing(
            func_2=self.compute_zeta_t,
            func_1=lambda: self.cls._zeta_t.value,
        )

    def test_eta_t(self):
        self.do_testing(
            func_2=self.compute_eta_t,
            func_1=lambda: self.cls._eta_t.value,
        )

    def test_properties(self) -> None:
        properties = ['x_0', 'norm_delta', 'kappa_t', 'mu']
        self._test_properties(properties)


# |x| << 1 << |x_v|
# |x_s| << 1 << |x_v_s|
# kappa_t << 1
# mu << 1

class TestARFLimit1Traveling(HelperTravelingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.small_viscous_boundary_layer = True
        self.large_viscous_boundary_layer = False
        self.small_particle_limit = True

    def compute_arf(self):
        out = 6 * pi * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= self.norm_delta * (self.rho_s - self.rho_f)**2
        out /= (1 + self.mu) * (2 * self.rho_s + self.rho_f)**2

        return out


# |x| << |x_v| << 1
# |x_s | << | x_v_s | << 1
# kappa_t << 1
# rho_f << rho_s

class TestARFLimit2Traveling(HelperTravelingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.small_viscous_boundary_layer = False
        self.large_viscous_boundary_layer = True
        self.small_particle_limit = True

    def compute_arf(self):
        out = pi / 6 * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= self.norm_delta**2
        out *= 5 - 2 * self.kappa_t - 8 * self.compute_f1()

        return out


class TestARFLimit2Standing(HelperStandingARF, TestARFBase):

    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.small_viscous_boundary_layer = False
        self.large_viscous_boundary_layer = True
        self.small_particle_limit = True

    def compute_arf(self):
        out = pi * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= self.compute_G() * sin(2 * self.position * self.cls.k_f.real)

        return out


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Union

import numpy as np

from osaft.core.functions import pi
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.plotting.grid import Grid
from osaft.solutions.base_scattering import BaseScattering
from osaft.solutions.base_streaming import BaseStreaming


class VelocityFieldData(ABC):
    """Container for plotting data for scattering and streaming plots

    :param sol: solution
    :param r_min: start point for radial plotting range in [m]
    :param r_max: end point for tangential plotting range  in [m]
    :param theta_min: start point for tangential plotting range in [rad]
    :param theta_max: end point for tangential plotting range in [rad]
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    """

    def __init__(
            self,
            sol: Union[BaseScattering, BaseStreaming],
            r_min: float,
            r_max: float,
            theta_min: float = 0,
            theta_max: float = pi,
            res: Union[int, tuple[int, int]] = 100,
    ) -> None:
        """Constructor method"""

        # Parameters for the coordinate grid
        self._r_min = PassiveVariable(r_min, 'r min')
        self._r_max = PassiveVariable(r_max, 'r max')
        self._theta_min = PassiveVariable(theta_min, 'theta min')
        self._theta_max = PassiveVariable(theta_max, 'r min')
        self._resolution = PassiveVariable(res, 'plotting resolution')

        # Solution to be plotted
        self._sol = PassiveVariable(sol, 'solution for plotting')
        self._mode = PassiveVariable(None, 'mode')

        # Dependent Variables: Data containers
        self._x = ActiveVariable(self._reset_real_array, 'x coordinate')
        self._z = ActiveVariable(self._reset_real_array, 'z coordinate')
        self._u = ActiveVariable(
            self._reset_complex_array, 'velocity/displacement '
            'in x-direction',
        )
        self._w = ActiveVariable(
            self._reset_complex_array, 'velocity/displacement '
            'in y-direction',
        )
        self._u_norm = ActiveVariable(
            self._reset_real_array, 'norm of the '
            'velocity / '
            'displacement',
        )
        # Dependent Variables: Coordinate grid
        self._grid = ActiveVariable(self._get_grid)

        # Setting links
        self._grid.is_computed_by(
            self._r_min, self._r_max, self._theta_min,
            self._theta_max, self._resolution,
        )
        self._x.is_computed_by(self._grid)
        self._w.is_computed_by(self._grid)
        self._u.is_computed_by(self._sol, self._grid, self._mode)
        self._w.is_computed_by(self._sol, self._grid, self._mode)
        self._u_norm.is_computed_by(self._sol, self._grid, self._mode)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_velocity_magnitude(
        self,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns the velocity magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity field at
        these points
        """
        return self.x, self.z, self.u_norm  # pragma: no cover

    def get_velocity_vector(self) -> tuple[
        np.ndarray, np.ndarray,
        np.ndarray, np.ndarray,
    ]:
        """Returns the velocity vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the velocity field vectors at
        these points
        """
        return self.x, self.z, self.u, self.w  # pragma: no cover

    # -------------------------------------------------------------------------
    # Getters and Setters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def sol(self) -> Union[BaseScattering, BaseStreaming]:
        """Solution for plotting

        :getter: returns the end point for radial plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._sol.value

    @sol.setter
    def sol(self, value: Union[BaseScattering, BaseStreaming]) -> None:
        self._sol.value = value

    @property
    def r_min(self) -> float:
        """End point for radial plotting range [m]

        :getter: returns the end point for radial plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._r_min.value

    @r_min.setter
    def r_min(self, value: float) -> None:
        self._r_min.value = value

    @property
    def r_max(self) -> float:
        """Start point for radial plotting range [m]

        :getter: returns the start point for radial plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._r_max.value

    @r_max.setter
    def r_max(self, value: float) -> None:
        self._r_max.value = value

    @property
    def theta_min(self) -> float:
        """Start point for tangential plotting range [rad]

        :getter: returns the start point for tangential plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._theta_min.value

    @theta_min.setter
    def theta_min(self, value: float) -> None:
        self._theta_min.value = value

    @property
    def theta_max(self) -> float:
        """Endpoint for tangential plotting range [rad]

        :getter: returns the end point for tangential plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._theta_max.value

    @theta_max.setter
    def theta_max(self, value: float) -> None:
        self._theta_max.value = value

    @property
    def resolution(self) -> Union[int, tuple[int, int]]:
        """Resolution for scattering / streaming plots

        `resolution` can be an `int` when the resolution in radial and
        tangential direction is the same, or a `tuple` of two `int`
        for independent resolutions (radial resolution, tangential_resolution)
        :getter: returns plotting resolution
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._resolution.value

    @resolution.setter
    def resolution(self, value: Union[int, tuple[int, int]]) -> None:
        self._resolution.value = value

    @property
    def mode(self) -> int:
        """Mode that is plotted. If `None` all modes up to `sol.N_max`
        are superimposed.
        :getter: returns the mode
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._mode.value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def grid(self) -> Grid:
        """Plotting grid
        """
        return self._grid.value

    @property
    def x(self) -> np.ndarray:
        """x-coordinates of the plotting grid in [m]
        """
        if np.all(self._x.value == 0):
            self._compute_data()
        return self._x.value

    @property
    def z(self) -> np.ndarray:
        """z-coordinates of the plotting grid in [m]
        """
        if np.all(self._z.value == 0):
            self._compute_data()  # pragma: no cover
        return self._z.value

    @property
    def u(self) -> np.ndarray:
        """Velocity / displacement in x-direction in [m/s]
        """
        if np.all(self._u.value == 0):
            self._compute_data()
        return self._u.value

    @property
    def w(self) -> np.ndarray:
        """Velocity / displacement in z-direction in [m/s]
        """
        if np.all(self._w.value == 0):
            self._compute_data()  # pragma: no cover
        return self._w.value

    @property
    def u_norm(self) -> np.ndarray:
        """Norm of the velocity / displacement in z-direction in [m/s]

        norm = abs(u) ** 2 + abs(w) ** 2
        """
        if np.all(self._u_norm.value == 0):
            self._compute_data()
        return self._u_norm.value

    # -------------------------------------------------------------------------
    # Methods for dependent variables
    # -------------------------------------------------------------------------

    def _reset_real_array(self):
        """Resets real arrays for coordinates and norm"""
        return np.zeros(
            self.grid.r_res * self.grid.theta_res,
            dtype=np.float64,
        )

    def _reset_complex_array(self):
        """Resets complex arrays for complex velocities and norm"""
        return np.zeros(
            self.grid.r_res * self.grid.theta_res,
            dtype=np.complex128,
        )

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _get_grid(self) -> Grid:
        """Returns a new coordinate grid"""
        return Grid(
            self.r_min, self.r_max,
            self.theta_min, self.theta_max,
            self.resolution,
        )

    @abstractmethod
    def _compute_data(
            self,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Compute the velocity data and store it in the respective attributes
        """
        pass  # pragma: no cover


if __name__ == '__main__':
    pass



import unittest

from osaft import Yosioka1955
from osaft.core.functions import pi
from osaft.plotting.datacontainers.scattering_datacontainer import (
    FluidScatteringData,
)
from osaft.solutions import Doinikov1994Rigid
from osaft.tests.basetest import BaseTest


class TestBaseScatteringDatacontainer(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.sol = Yosioka1955.ScatteringField(
            self.f, self.R_0,
            self.rho_s, self.c_s,
            self.rho_f, self.c_f,
            self.p_0,
            self.wave_type,
            self.position,
        )

        self.cls = FluidScatteringData(
            self.sol, 5 * self.sol.R_0,
        )

    def test_setters_and_getters(self):
        # Change solution
        sol_2 = Doinikov1994Rigid.ScatteringField(
            self.f, self.R_0,
            self.rho_s,
            self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )
        self.cls.sol = sol_2

        # Change r_min
        self.cls.r_min = 2e-6
        self.assertEqual(self.cls.r_min, 2e-6)

        # Change r_max
        self.cls.r_max = 10e-6
        self.assertEqual(self.cls.r_max, 10e-6)

        # Change theta min
        self.cls.theta_min = pi / 4
        self.assertEqual(self.cls.theta_min, pi / 4)

        # Change theta max
        self.cls.theta_max = pi / 2
        self.assertEqual(self.cls.theta_max, pi / 2)

    def test_change_resolution(self):
        self.cls.resolution = 200
        self.assertEqual(self.cls.grid.r_res, 200)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from osaft import Doinikov1994Rigid
from osaft.core.functions import pi
from osaft.core.functions import spherical_2_cartesian_coordinates as s2c_coord
from osaft.core.functions import spherical_2_cartesian_vector as s2c_vector
from osaft.plotting.datacontainers.wireframe_datacontainer import (
    DeformedLine,
    ParticleWireframeData,
)
from osaft.tests.basetest import BaseTest


class TestWireframeDatacontainer(BaseTest):

    def setUp(self) -> None:

        super().setUp()

        self.sol = Doinikov1994Rigid.ScatteringField(
            self.f, self.R_0,
            self.rho_s,
            self.rho_f, self.c_f,
            self.eta_f, self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )

        self.cls = ParticleWireframeData(
            self.sol, nbr_r_levels=10,
            nbr_theta_levels=10,
            resolution=(50, 50),
            scale_factor=0.2,
        )

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def arr_r(self):
        return np.linspace(1e-30, self.sol.R_0, self.cls.resolution[0])

    def arr_theta(self):
        if self.cls.symmetric:
            return np.linspace(0, pi, self.cls.resolution[1])
        else:
            return np.linspace(0, 2 * pi, self.cls.resolution[1])

    def deformed_radius(self, theta):
        # Displacements
        u_r = self.sol.radial_particle_displacement(
            self.arr_r(), theta, t=0,
        )
        u_t = self.sol.tangential_particle_displacement(
            self.arr_r(), theta, t=0,
        )
        u, v = s2c_vector(u_r, u_t, theta)
        u_scaled = u * self.cls.scale_factor
        v_scaled = v * self.cls.scale_factor

        # Coordinates
        x, y = s2c_coord(self.arr_r(), theta)

        return DeformedLine(x, y, u_scaled, v_scaled)

    def deformed_circle(self, radius):
        # Displacements
        u_r = self.sol.radial_particle_displacement(
            radius, self.arr_theta(), t=0,
        )
        u_t = self.sol.tangential_particle_displacement(
            radius, self.arr_theta(), t=0,
        )
        u, v = s2c_vector(u_r, u_t, self.arr_theta())
        u_scaled = u * self.cls.scale_factor
        v_scaled = v * self.cls.scale_factor

        # Coordinates
        x, y = s2c_coord(radius, self.arr_theta())

        return DeformedLine(x, y, u_scaled, v_scaled)

    # -------------------------------------------------------------------------
    # Helpers
    # -------------------------------------------------------------------------

    @staticmethod
    def compare_deformed_lines(df_1: DeformedLine, df_2: DeformedLine):
        x_1, y_1 = df_1.get_reference()
        x_2, y_2 = df_2.get_reference()
        u_1, v_1 = df_2.get_deformed(0)
        u_2, v_2 = df_2.get_deformed(0)
        np.testing.assert_array_equal(x_1, x_2)
        np.testing.assert_array_equal(y_1, y_2)
        np.testing.assert_array_equal(u_1, u_2)
        np.testing.assert_array_equal(v_1, v_2)

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_nbr_t_level_setter(self) -> None:
        self.cls.symmetric = True
        self.cls.nbr_theta_levels = 9
        self.assertEqual(self.cls.nbr_theta_levels, 10)
        self.cls.symmetric = False
        self.cls.nbr_theta_levels = 9
        self.assertEqual(self.cls.nbr_theta_levels, 9)

    def test_nbr_levels_dependencies(self) -> None:
        self.cls.nbr_r_levels = 15
        self.assertEqual(len(self.cls.r_levels), 15)
        self.cls.nbr_theta_levels = 16
        self.assertEqual(len(self.cls.theta_levels), 8)

    def test_resolution_setter(self) -> None:
        def test_res(value: int) -> None:
            self.assertEqual(self.cls.resolution[0], value)
        self.cls.resolution = (100, 100)
        test_res(100)
        self.cls.resolution = 120
        test_res(120)
        self.cls.resolution = [130]
        test_res(130)
        with self.assertRaises(ValueError):
            self.cls.resolution = (100, 100, 100)

    def test_symmetric(self) -> None:
        self.cls.nbr_theta_levels = 16
        self.assertEqual(len(self.cls.theta_levels), 8)
        self.cls.symmetric = False
        self.assertEqual(len(self.cls.theta_levels), 16)

    def test_scale_factor(self) -> None:
        self.cls.rel_scale_factor = 1
        sf_1 = self.cls.scale_factor
        self.cls.rel_scale_factor = 2
        sf_2 = self.cls.scale_factor
        self.assertAlmostEqual(sf_1, sf_2 / 2)

    def test_setters_and_getters(self) -> None:
        # Solution
        self.cls.sol = self.sol
        self.assertEqual(self.cls.sol, self.sol)
        # nbr_r_levels
        self.cls.nbr_r_levels = 10
        self.assertEqual(self.cls.nbr_r_levels, 10)
        # Mode
        self.cls.mode = None
        self.assertIsNone(self.cls.mode)
        # Scale Factor
        self.cls.rel_scale_factor = 0.3
        self.assertEqual(self.cls.rel_scale_factor, 0.3)

    def test_arr_r(self) -> None:
        np.testing.assert_array_equal(self.arr_r(), self.cls.arr_r)
        self.assertEqual(self.cls.arr_r.max(), self.sol.R_0)

    def test_arr_theta(self):
        # Not symmetric
        self.cls.symmetric = False
        np.testing.assert_array_equal(self.arr_theta(), self.cls.arr_theta)
        self.assertEqual(self.cls.arr_theta.max(), 2 * pi)
        # Symmetric
        self.cls.symmetric = True
        np.testing.assert_array_equal(self.arr_theta(), self.cls.arr_theta)
        self.assertEqual(self.cls.arr_theta.max(), pi)

    def test_deformed_circles(self):
        deformed_circles = self.cls.deformed_circles
        for index, radius in enumerate(self.cls.r_levels):
            df_1 = deformed_circles[index]
            df_2 = self.deformed_circle(radius)
            self.compare_deformed_lines(df_1, df_2)

    def test_deformed_radii(self):
        deformed_radii, _ = self.cls.get_displacements()
        for index, theta in enumerate(self.cls.theta_levels):
            df_1 = deformed_radii[index]
            df_2 = self.deformed_radius(theta)
            self.compare_deformed_lines(df_1, df_2)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from osaft import King1934
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import SpecialFunctions as Sp
from osaft.core.functions import cos, exp, sin
from osaft.tests.basetest_scattering import HelperScattering
from osaft.tests.King1934.test_base import BaseKing


class TestScattering(HelperScattering, BaseKing):

    def setUp(self) -> None:

        super().setUp()

        self.small_particle_limit = False
        self.cls = King1934.ScatteringField(
            self.f,
            self.R_0, self.rho_s,
            self.rho_f, self.c_f,
            self.p_0, self.wave_type,
            self.position,
        )
        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def A_dash_n(self, n: int):
        arg = self.cls.alpha
        out = - self.cls.A_in(n) * self.G_n(n, arg)
        out /= (self.G_n(n, arg) + 1j * self.F_n(n, arg))
        return out

    def V_r_sc(self, n, r) -> complex:
        out = self.A_dash_n(n)
        out *= self.cls.k_f
        out *= Sp.d1_hankelh2(n, self.cls.k_f * r)
        return out

    def V_theta_sc(self, n, r) -> complex:
        out = self.A_dash_n(n)
        out *= Sp.hankelh2(n, self.cls.k_f * r)
        out /= r
        return out

    def Phi_incident(
        self, r, theta, t,
    ) -> complex:
        def Phi_n(n):
            arg = self.cls.A_in(n) * Sp.besselj(n, r * self.cls.k_f)
            out = exp(-1j * self.cls.omega * t)
            out *= Leg.cos_monomial(n, theta, arg)
            return out

        return np.sum([Phi_n(n) for n in range(self.N_max + 1)])

    def Phi_scattering(
        self, r, theta, t,
    ) -> complex:
        def Phi_n(n):
            arg = - self.G_n(n, self.cls.alpha)
            arg /= (
                self.G_n(n, self.cls.alpha) + 1j * self.F_n(n, self.cls.alpha)
            )
            arg *= self.cls.A_in(n) * Sp.hankelh2(n, r * self.cls.k_f)
            out = exp(-1j * self.cls.omega * t)
            out *= Leg.cos_monomial(n, theta, arg)
            return out

        return np.sum([Phi_n(n) for n in range(self.N_max + 1)])

    def radial_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return self.particle_velocity(t) * cos(theta)
        else:
            return 0

    def tangential_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return -self.particle_velocity(t) * sin(theta)
        else:
            return 0

    def particle_velocity(self, t):
        alpha = self.cls.alpha

        out = self.F_n(1, alpha)
        out -= 1j * self.G_n(1, alpha)
        out = 1 / out

        out *= self.cls.A_in(1) / alpha**3
        out *= self.cls.k_f * self.rho_f / self.rho_s
        out *= exp(-1j * self.cls.omega * t)

        return out

    # -------------------------------------------------------------------------
    # Test coefficients
    # -------------------------------------------------------------------------

    def test_methods(self) -> None:
        method_dict = {}
        method_dict['A_dash_n'] = None

        # need to be float otherwise
        # Integers to negative integer powers are not allowed.
        method_dict['phi_n'] = 2.0
        method_dict['psi_n'] = 2.0

        arg = self.cls.k_f * self.get_random_r(self.cls)
        method_dict['F_n'] = arg
        method_dict['G_n'] = arg

        arg = self.get_random_r(self.cls)
        method_dict['V_r_sc'] = arg
        method_dict['V_theta_sc'] = arg

        self._test_methods_n(method_dict)

    def test_Phi_incident(self) -> None:
        for i in range(5):
            r = self.get_random_r(self.cls)
            t = self.get_random_t()
            theta = self.get_random_theta()
            self.N_max = i
            self.cls.N_max = i

            self.do_testing(
                func_1=self.Phi_incident, args_1=(r, theta, t),
                func_2=self.cls.Phi_incident, args_2=(r, theta, t),
            )

    def test_Phi_scattering(self) -> None:
        for i in range(5):
            r = self.get_random_r(self.cls)
            t = self.get_random_t()
            theta = self.get_random_theta()

            self.N_max = i
            self.cls.N_max = i

            self.do_testing(
                func_1=self.Phi_scattering, args_1=(r, theta, t),
                func_2=self.cls.Phi_scattering, args_2=(r, theta, t),
            )


if __name__ == '__main__':
    unittest.main()



from collections import Sequence
from typing import Union

import numpy as np

from osaft.core.functions import pi


class Grid:
    """Data grid class for plotting

    Creates a data grid in spherical coordinates between `r_min` and `r_max`,
    and `theta_min` and `theta_max`. If `outside` `True` the grid is adjusted
    for plotting the fluid field in a square box with side `r_max`.
    Therefore, the grid goes from `r_min` to `sqrt(2) * r_max`


    :param r_min: start point for radial plotting range in [m]
    :param r_max: end point for tangential plotting range  in [m]
    :param theta_min: start point for tangential plotting range in [rad]
    :param theta_max: end point for tangential plotting range in [rad]
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    """
    def __init__(
        self,
        r_min: float,
        r_max: float,
        theta_min: float = 0,
        theta_max: float = pi,
        res: Union[int, tuple[int, int]] = 100,
    ) -> None:
        """Constructor method"""
        self.r_min = r_min
        self.r_max = r_max
        self.theta_min = theta_min
        self.theta_max = theta_max
        self.r_res, self.theta_res = self._unpack_resolution(res)

    # -------------------------------------------------------------------------
    # Init Helper Methods
    # -------------------------------------------------------------------------

    @staticmethod
    def _unpack_resolution(res: Union[int, tuple[int, int]]):
        """ Unpacks resolution tuple if needed

        if `res` is an `int` a tuple `(res, res)` is
        returned. If `res` is a tuple with two values, `res` is passed through.
        :param res: res of the grid
        """
        if isinstance(res, Sequence):
            if len(res) == 1:
                return res[0], res[0]
            elif len(res) == 2:
                return res
            else:
                raise ValueError(
                    'Resolution needs to be either one value for both '
                    'radial and tangential direction',
                )
        else:
            return res, res

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def dr(self):
        """Step size in radial direction"""
        return (self.r_max - self.r_min) / self.r_res

    @property
    def dtheta(self):
        """Step size in tangential direction"""
        return (self.theta_max - self.theta_min) / self.theta_res

    @property
    def arr_r(self):
        """Linearly spaced array in radial direction"""
        return np.linspace(self.r_min, self.r_max, self.r_res)

    @property
    def arr_theta(self):
        """Linearly spaced array in tangential direction"""
        return np.linspace(self.theta_min, self.theta_max, self.theta_res)


if __name__ == '__main__':
    pass
